# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
# pyre-unsafe

from typing import Any, List, Optional, Sequence, Union

from libcst._exceptions import PartialParserSyntaxError
from libcst._maybe_sentinel import MaybeSentinel
from libcst._nodes.expression import (
    Annotation,
    Name,
    Param,
    Parameters,
    ParamSlash,
    ParamStar,
)
from libcst._nodes.op import AssignEqual, Comma
from libcst._parser.custom_itertools import grouper
from libcst._parser.production_decorator import with_production
from libcst._parser.types.config import ParserConfig
from libcst._parser.types.partials import ParamStarPartial
from libcst._parser.whitespace_parser import parse_parenthesizable_whitespace


@with_production(  # noqa: C901: too complex
    "typedargslist",
    """(
      (tfpdef_assign (',' tfpdef_assign)* ',' tfpdef_posind [',' [ tfpdef_assign (
            ',' tfpdef_assign)* [',' [
            tfpdef_star (',' tfpdef_assign)* [',' [tfpdef_starstar [',']]]
          | tfpdef_starstar [',']]]
      | tfpdef_star (',' tfpdef_assign)* [',' [tfpdef_starstar [',']]]
      | tfpdef_starstar [',']]] )
    |  (tfpdef_assign (',' tfpdef_assign)* [',' [
       tfpdef_star (',' tfpdef_assign)* [',' [tfpdef_starstar [',']]]
      | tfpdef_starstar [',']]]
      | tfpdef_star (',' tfpdef_assign)* [',' [tfpdef_starstar [',']]]
      | tfpdef_starstar [','])
    )""",
    version=">=3.8",
)
@with_production(  # noqa: C901: too complex
    "typedargslist",
    (
        "(tfpdef_assign (',' tfpdef_assign)* "
        + "[',' [tfpdef_star (',' tfpdef_assign)* [',' [tfpdef_starstar [',']]] | tfpdef_starstar [',']]]"
        + "| tfpdef_star (',' tfpdef_assign)* [',' [tfpdef_starstar [',']]] | tfpdef_starstar [','])"
    ),
    version=">=3.6,<=3.7",
)
@with_production(  # noqa: C901: too complex
    "typedargslist",
    (
        "(tfpdef_assign (',' tfpdef_assign)* "
        + "[',' [tfpdef_star (',' tfpdef_assign)* [',' tfpdef_starstar] | tfpdef_starstar]]"
        + "| tfpdef_star (',' tfpdef_assign)* [',' tfpdef_starstar] | tfpdef_starstar)"
    ),
    version="<=3.5",
)
@with_production(
    "varargslist",
    """vfpdef_assign (',' vfpdef_assign)* ',' vfpdef_posind [',' [ (vfpdef_assign (',' vfpdef_assign)* [',' [
            vfpdef_star (',' vfpdef_assign)* [',' [vfpdef_starstar [',']]]
          | vfpdef_starstar [',']]]
      | vfpdef_star (',' vfpdef_assign)* [',' [vfpdef_starstar [',']]]
      | vfpdef_starstar [',']) ]] | (vfpdef_assign (',' vfpdef_assign)* [',' [
            vfpdef_star (',' vfpdef_assign)* [',' [vfpdef_starstar [',']]]
          | vfpdef_starstar [',']]]
      | vfpdef_star (',' vfpdef_assign)* [',' [vfpdef_starstar [',']]]
      | vfpdef_starstar [',']
    )""",
    version=">=3.8",
)
@with_production(
    "varargslist",
    (
        "(vfpdef_assign (',' vfpdef_assign)* "
        + "[',' [vfpdef_star (',' vfpdef_assign)* [',' [vfpdef_starstar [',']]] | vfpdef_starstar [',']]]"
        + "| vfpdef_star (',' vfpdef_assign)* [',' [vfpdef_starstar [',']]] | vfpdef_starstar [','])"
    ),
    version=">=3.6,<=3.7",
)
@with_production(
    "varargslist",
    (
        "(vfpdef_assign (',' vfpdef_assign)* "
        + "[',' [vfpdef_star (',' vfpdef_assign)* [',' vfpdef_starstar] | vfpdef_starstar]]"
        + "| vfpdef_star (',' vfpdef_assign)* [',' vfpdef_starstar] | vfpdef_starstar)"
    ),
    version="<=3.5",
)
def convert_argslist(  # noqa: C901
    config: ParserConfig, children: Sequence[Any]
) -> Any:
    posonly_params: List[Param] = []
    posonly_ind: Union[ParamSlash, MaybeSentinel] = MaybeSentinel.DEFAULT
    params: List[Param] = []
    seen_default: bool = False
    star_arg: Union[Param, ParamStar, MaybeSentinel] = MaybeSentinel.DEFAULT
    kwonly_params: List[Param] = []
    star_kwarg: Optional[Param] = None

    def add_param(
        current_param: Optional[List[Param]], param: Union[Param, ParamStar]
    ) -> Optional[List[Param]]:
        nonlocal star_arg
        nonlocal star_kwarg
        nonlocal seen_default
        nonlocal posonly_params
        nonlocal posonly_ind
        nonlocal params

        if isinstance(param, ParamStar):
            # Only can add this if we don't already have a "*" or a "*param".
            if current_param is params:
                star_arg = param
                current_param = kwonly_params
            else:
                # Example code:
                #     def fn(*abc, *): ...
                # This should be unreachable, the grammar already disallows it.
                raise Exception(
                    "Cannot have multiple star ('*') markers in a single argument "
                    + "list."
                )
        elif isinstance(param, ParamSlash):
            # Only can add this if we don't already have a "/" or a "*" or a "*param".
            if current_param is params and len(posonly_params) == 0:
                posonly_ind = param
                posonly_params = params
                params = []
                current_param = params
            else:
                # Example code:
                # def fn(foo, /, *, /, bar): ...
                # This should be unreachable, the grammar already disallows it.
                raise Exception(
                    "Cannot have multiple slash ('/') markers in a single argument "
                    + "list."
                )
        elif isinstance(param.star, str) and param.star == "" and param.default is None:
            # Can only add this if we're in the params or kwonly_params section
            if current_param is params and not seen_default:
                params.append(param)
            elif current_param is kwonly_params:
                kwonly_params.append(param)
            else:
                # Example code:
                #     def fn(first=None, second): ...
                # This code is reachable, so we should use a PartialParserSyntaxError.
                raise PartialParserSyntaxError(
                    "Cannot have a non-default argument following a default argument."
                )
        elif (
            isinstance(param.star, str)
            and param.star == ""
            and param.default is not None
        ):
            # Can only add this if we're not yet at star args.
            if current_param is params:
                seen_default = True
                params.append(param)
            elif current_param is kwonly_params:
                kwonly_params.append(param)
            else:
                # Example code:
                #     def fn(**kwargs, trailing=None)
                # This should be unreachable, the grammar already disallows it.
                raise Exception("Cannot have any arguments after a kwargs expansion.")
        elif (
            isinstance(param.star, str) and param.star == "*" and param.default is None
        ):
            # Can only add this if we're in params, since we only allow one of
            # "*" or "*param".
            if current_param is params:
                star_arg = param
                current_param = kwonly_params
            else:
                # Example code:
                #     def fn(*first, *second): ...
                # This should be unreachable, the grammar already disallows it.
                raise Exception(
                    "Expected a keyword argument but found a starred positional "
                    + "argument expansion."
                )
        elif (
            isinstance(param.star, str) and param.star == "**" and param.default is None
        ):
            # Can add this in all cases where we don't have a star_kwarg
            # yet.
            if current_param is not None:
                star_kwarg = param
                current_param = None
            else:
                # Example code:
                #     def fn(**first, **second)
                # This should be unreachable, the grammar already disallows it.
                raise Exception(
                    "Multiple starred keyword argument expansions are not allowed in a "
                    + "single argument list"
                )
        else:
            # The state machine should never end up here.
            raise Exception("Logic error!")

        return current_param

    # The parameter list we are adding to
    current: Optional[List[Param]] = params

    # We should have every other item in the group as a param or a comma by now,
    # so split them up, add commas and then put them in the appropriate group.
    for parameter, comma in grouper(children, 2):
        if comma is None:
            if isinstance(parameter, ParamStarPartial):
                # Example:
                #     def fn(abc, *): ...
                #
                # There's also the case where we have bare * with a trailing comma.
                # That's handled later.
                #
                # It's not valid to construct a ParamStar object without a comma, so we
                # need to catch the non-comma case separately.
                raise PartialParserSyntaxError(
                    "Named (keyword) arguments must follow a bare *."
                )
            else:
                current = add_param(current, parameter)
        else:
            comma = Comma(
                whitespace_before=parse_parenthesizable_whitespace(
                    config, comma.whitespace_before
                ),
                whitespace_after=parse_parenthesizable_whitespace(
                    config, comma.whitespace_after
                ),
            )
            if isinstance(parameter, ParamStarPartial):
                current = add_param(current, ParamStar(comma=comma))
            else:
                current = add_param(current, parameter.with_changes(comma=comma))

    if isinstance(star_arg, ParamStar) and len(kwonly_params) == 0:
        # Example:
        #     def fn(abc, *,): ...
        #
        # This will raise a validation error, but we want to make sure to raise a syntax
        # error instead.
        #
        # The case where there's no trailing comma is already handled by this point, so
        # this conditional is only for the case where we have a trailing comma.
        raise PartialParserSyntaxError(
            "Named (keyword) arguments must follow a bare *."
        )

    return Parameters(
        posonly_params=tuple(posonly_params),
        posonly_ind=posonly_ind,
        params=tuple(params),
        star_arg=star_arg,
        kwonly_params=tuple(kwonly_params),
        star_kwarg=star_kwarg,
    )


@with_production("tfpdef_star", "'*' [tfpdef]")
@with_production("vfpdef_star", "'*' [vfpdef]")
def convert_fpdef_star(config: ParserConfig, children: Sequence[Any]) -> Any:
    if len(children) == 1:
        (star,) = children
        return ParamStarPartial()
    else:
        star, param = children
        return param.with_changes(
            star=star.string,
            whitespace_after_star=parse_parenthesizable_whitespace(
                config, star.whitespace_after
            ),
        )


@with_production("tfpdef_starstar", "'**' tfpdef")
@with_production("vfpdef_starstar", "'**' vfpdef")
def convert_fpdef_starstar(config: ParserConfig, children: Sequence[Any]) -> Any:
    starstar, param = children
    return param.with_changes(
        star=starstar.string,
        whitespace_after_star=parse_parenthesizable_whitespace(
            config, starstar.whitespace_after
        ),
    )


@with_production("tfpdef_assign", "tfpdef ['=' test]")
@with_production("vfpdef_assign", "vfpdef ['=' test]")
def convert_fpdef_assign(config: ParserConfig, children: Sequence[Any]) -> Any:
    if len(children) == 1:
        (child,) = children
        return child

    param, equal, default = children
    return param.with_changes(
        equal=AssignEqual(
            whitespace_before=parse_parenthesizable_whitespace(
                config, equal.whitespace_before
            ),
            whitespace_after=parse_parenthesizable_whitespace(
                config, equal.whitespace_after
            ),
        ),
        default=default.value,
    )


@with_production("tfpdef", "NAME [':' test]")
@with_production("vfpdef", "NAME")
def convert_fpdef(config: ParserConfig, children: Sequence[Any]) -> Any:
    if len(children) == 1:
        # This is just a parameter
        (child,) = children
        namenode = Name(child.string)
        annotation = None
    else:
        # This is a parameter with a type hint
        name, colon, typehint = children
        namenode = Name(name.string)
        annotation = Annotation(
            whitespace_before_indicator=parse_parenthesizable_whitespace(
                config, colon.whitespace_before
            ),
            whitespace_after_indicator=parse_parenthesizable_whitespace(
                config, colon.whitespace_after
            ),
            annotation=typehint.value,
        )

    return Param(star="", name=namenode, annotation=annotation, default=None)


@with_production("tfpdef_posind", "'/'")
@with_production("vfpdef_posind", "'/'")
def convert_fpdef_slash(config: ParserConfig, children: Sequence[Any]) -> Any:
    return ParamSlash()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import codecs
import re
import sys
from dataclasses import dataclass, field, fields
from enum import Enum
from typing import Any, Callable, FrozenSet, List, Mapping, Optional, Pattern, Union

from libcst._add_slots import add_slots
from libcst._nodes.whitespace import NEWLINE_RE
from libcst._parser.parso.utils import parse_version_string, PythonVersionInfo

_INDENT_RE: Pattern[str] = re.compile(r"[ \t]+")

try:
    from libcst_native import parser_config as config_mod

    MockWhitespaceParserConfig = config_mod.BaseWhitespaceParserConfig
except ImportError:
    from libcst._parser.types import py_config as config_mod

    MockWhitespaceParserConfig = config_mod.MockWhitespaceParserConfig

BaseWhitespaceParserConfig = config_mod.BaseWhitespaceParserConfig
ParserConfig = config_mod.ParserConfig
parser_config_asdict: Callable[
    [ParserConfig], Mapping[str, Any]
] = config_mod.parser_config_asdict


class AutoConfig(Enum):
    """
    A sentinel value used in PartialParserConfig
    """

    token: int = 0

    def __repr__(self) -> str:
        return str(self)


# This list should be kept in sorted order.
KNOWN_PYTHON_VERSION_STRINGS = ["3.0", "3.1", "3.3", "3.5", "3.6", "3.7", "3.8"]


@add_slots
@dataclass(frozen=True)
class PartialParserConfig:
    r"""
    An optional object that can be supplied to the parser entrypoints (e.g.
    :func:`parse_module`) to configure the parser.

    Unspecified fields will be inferred from the input source code or from the execution
    environment.

    >>> import libcst as cst
    >>> tree = cst.parse_module("abc")
    >>> tree.bytes
    b'abc'
    >>> # override the default utf-8 encoding
    ... tree = cst.parse_module("abc", cst.PartialParserConfig(encoding="utf-32"))
    >>> tree.bytes
    b'\xff\xfe\x00\x00a\x00\x00\x00b\x00\x00\x00c\x00\x00\x00'
    """

    #: The version of Python that the input source code is expected to be syntactically
    #: compatible with. This may be different from the Python interpreter being used to
    #: run LibCST. For example, you can parse code as 3.7 with a CPython 3.6
    #: interpreter.
    #:
    #: If unspecified, it will default to the syntax of the running interpreter
    #: (rounding down from among the following list).
    #:
    #: Currently, only Python 3.0, 3.1, 3.3, 3.5, 3.6, 3.7 and 3.8 syntax is supported.
    #: The gaps did not have any syntax changes from the version prior.
    python_version: Union[str, AutoConfig] = AutoConfig.token

    #: A named tuple with the ``major`` and ``minor`` Python version numbers. This is
    #: derived from :attr:`python_version` and should not be supplied to the
    #: :class:`PartialParserConfig` constructor.
    parsed_python_version: PythonVersionInfo = field(init=False)

    #: The file's encoding format. When parsing a ``bytes`` object, this value may be
    #: inferred from the contents of the parsed source code. When parsing a ``str``,
    #: this value defaults to ``"utf-8"``.
    encoding: Union[str, AutoConfig] = AutoConfig.token

    #: Detected ``__future__`` import names
    future_imports: Union[FrozenSet[str], AutoConfig] = AutoConfig.token

    #: The indentation of the file, expressed as a series of tabs and/or spaces. This
    #: value is inferred from the contents of the parsed source code by default.
    default_indent: Union[str, AutoConfig] = AutoConfig.token

    #: The newline of the file, expressed as ``\n``, ``\r\n``, or ``\r``. This value is
    #: inferred from the contents of the parsed source code by default.
    default_newline: Union[str, AutoConfig] = AutoConfig.token

    def __post_init__(self) -> None:
        raw_python_version = self.python_version

        if isinstance(raw_python_version, AutoConfig):
            # If unspecified, we'll try to pick the same as the running
            # interpreter.  There will always be at least one entry.
            parsed_python_version = _pick_compatible_python_version()
        else:
            # If the caller specified a version, we require that to be a known
            # version (because we don't want to encourage doing duplicate work
            # when there weren't syntax changes).

            # `parse_version_string` will raise a ValueError if the version is
            # invalid.
            parsed_python_version = parse_version_string(raw_python_version)

        if not any(
            parsed_python_version == parse_version_string(v)
            for v in KNOWN_PYTHON_VERSION_STRINGS
        ):
            comma_versions = ", ".join(KNOWN_PYTHON_VERSION_STRINGS)
            raise ValueError(
                "LibCST can only parse code using one of the following versions of "
                + f"Python's grammar: {comma_versions}. More versions may be "
                + "supported by future releases."
            )

        # We use object.__setattr__ because the dataclass is frozen. See:
        # https://docs.python.org/3/library/dataclasses.html#frozen-instances
        # This should be safe behavior inside of `__post_init__`.
        object.__setattr__(self, "parsed_python_version", parsed_python_version)

        encoding = self.encoding
        if not isinstance(encoding, AutoConfig):
            try:
                codecs.lookup(encoding)
            except LookupError:
                raise ValueError(f"{repr(encoding)} is not a supported encoding")

        newline = self.default_newline
        if (
            not isinstance(newline, AutoConfig)
            and NEWLINE_RE.fullmatch(newline) is None
        ):
            raise ValueError(
                f"Got an invalid value for default_newline: {repr(newline)}"
            )

        indent = self.default_indent
        if not isinstance(indent, AutoConfig) and _INDENT_RE.fullmatch(indent) is None:
            raise ValueError(f"Got an invalid value for default_indent: {repr(indent)}")

    def __repr__(self) -> str:
        init_keys: List[str] = []

        for f in fields(self):
            # We don't display the parsed_python_version attribute because it contains
            # the same value as python_version, only parsed.
            if f.name == "parsed_python_version":
                continue
            value = getattr(self, f.name)
            if not isinstance(value, AutoConfig):
                init_keys.append(f"{f.name}={value!r}")

        return f"{self.__class__.__name__}({', '.join(init_keys)})"


def _pick_compatible_python_version(version: Optional[str] = None) -> PythonVersionInfo:
    max_version = parse_version_string(version)
    for v in KNOWN_PYTHON_VERSION_STRINGS[::-1]:
        tmp = parse_version_string(v)
        if tmp <= max_version:
            return tmp

    raise ValueError(
        f"No version found older than {version} ({max_version}) while "
        + f"running on {sys.version_info}"
    )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Optional, Sequence, Union

from libcst._nodes.whitespace import (
    EmptyLine,
    Newline,
    ParenthesizedWhitespace,
    SimpleWhitespace,
    TrailingWhitespace,
)
from libcst._parser.types.config import BaseWhitespaceParserConfig as Config
from libcst._parser.types.whitespace_state import WhitespaceState as State

def parse_simple_whitespace(config: Config, state: State) -> SimpleWhitespace: ...
def parse_empty_lines(
    config: Config,
    state: State,
    *,
    override_absolute_indent: Optional[str] = None,
) -> Sequence[EmptyLine]: ...
def parse_trailing_whitespace(config: Config, state: State) -> TrailingWhitespace: ...
def parse_parenthesizable_whitespace(
    config: Config, state: State
) -> Union[SimpleWhitespace, ParenthesizedWhitespace]: ...



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
from libcst import MetadataWrapper, parse_module
from libcst.codemod import CodemodContext, CodemodTest
from libcst.codemod.visitors import GatherNamesFromStringAnnotationsVisitor
from libcst.testing.utils import UnitTest


class TestGatherNamesFromStringAnnotationsVisitor(UnitTest):
    def gather_names(self, code: str) -> GatherNamesFromStringAnnotationsVisitor:
        mod = MetadataWrapper(parse_module(CodemodTest.make_fixture_data(code)))
        mod.resolve_many(GatherNamesFromStringAnnotationsVisitor.METADATA_DEPENDENCIES)
        instance = GatherNamesFromStringAnnotationsVisitor(CodemodContext(wrapper=mod))
        mod.visit(instance)
        return instance

    def test_no_annotations(self) -> None:
        visitor = self.gather_names(
            """
            def foo() -> None:
                pass
            """
        )
        self.assertEqual(visitor.names, set())

    def test_simple_string_annotations(self) -> None:
        visitor = self.gather_names(
            """
            def foo() -> "None":
                pass
            """
        )
        self.assertEqual(visitor.names, {"None"})

    def test_concatenated_string_annotations(self) -> None:
        visitor = self.gather_names(
            """
            def foo() -> "No" "ne":
                pass
            """
        )
        self.assertEqual(visitor.names, {"None"})

    def test_typevars(self) -> None:
        visitor = self.gather_names(
            """
            from typing import TypeVar as SneakyBastard
            V = SneakyBastard("V", bound="int")
            """
        )
        self.assertEqual(visitor.names, {"V", "int"})

    def test_complex(self) -> None:
        visitor = self.gather_names(
            """
            from typing import TypeVar, TYPE_CHECKING
            if TYPE_CHECKING:
                from a import Container, Item
            def foo(a: "A") -> "Item":
                pass
            A = TypeVar("A", bound="Container[Item]")
            class X:
                var: "ThisIsExpensiveToImport"  # noqa
            """
        )
        self.assertEqual(
            visitor.names, {"A", "Item", "Container", "ThisIsExpensiveToImport"}
        )

    def test_dotted_names(self) -> None:
        visitor = self.gather_names(
            """
            a: "api.http_exceptions.HttpException"
            """
        )
        self.assertEqual(
            visitor.names,
            {"api", "api.http_exceptions", "api.http_exceptions.HttpException"},
        )



from token import (
    AMPER,
    AMPEREQUAL,
    AT,
    ATEQUAL,
    CIRCUMFLEX,
    CIRCUMFLEXEQUAL,
    COLON,
    COLONEQUAL,
    COMMA,
    COMMENT,
    DEDENT,
    DOT,
    DOUBLESLASH,
    DOUBLESLASHEQUAL,
    DOUBLESTAR,
    DOUBLESTAREQUAL,
    ELLIPSIS,
    ENCODING,
    ENDMARKER,
    EQEQUAL,
    EQUAL,
    ERRORTOKEN,
    EXACT_TOKEN_TYPES,
    GREATER,
    GREATEREQUAL,
    INDENT,
    LBRACE,
    LEFTSHIFT,
    LEFTSHIFTEQUAL,
    LESS,
    LESSEQUAL,
    LPAR,
    LSQB,
    MINEQUAL,
    MINUS,
    N_TOKENS,
    NAME,
    NEWLINE,
    NL,
    NOTEQUAL,
    NT_OFFSET,
    NUMBER,
    OP,
    PERCENT,
    PERCENTEQUAL,
    PLUS,
    PLUSEQUAL,
    RARROW,
    RBRACE,
    RIGHTSHIFT,
    RIGHTSHIFTEQUAL,
    RPAR,
    RSQB,
    SEMI,
    SLASH,
    SLASHEQUAL,
    STAR,
    STAREQUAL,
    STRING,
    TILDE,
    TYPE_COMMENT,
    TYPE_IGNORE,
    VBAR,
    VBAREQUAL,
)
from typing import Callable, Generator, Sequence, Tuple

Hexnumber: str = ...
Binnumber: str = ...
Octnumber: str = ...
Decnumber: str = ...
Intnumber: str = ...
Exponent: str = ...
Pointfloat: str = ...
Expfloat: str = ...
Floatnumber: str = ...
Imagnumber: str = ...
Number: str = ...
Whitespace: str = ...
Comment: str = ...
Ignore: str = ...
Name: str = ...

class TokenInfo(Tuple[int, str, Tuple[int, int], Tuple[int, int], int]):
    exact_type: int = ...
    type: int = ...
    string: str = ...
    start: Tuple[int, int] = ...
    end: Tuple[int, int] = ...
    line: int = ...
    def __repr__(self) -> str: ...

def detect_encoding(readline: Callable[[], bytes]) -> Tuple[str, Sequence[bytes]]: ...
def tokenize(readline: Callable[[], bytes]) -> Generator[TokenInfo, None, None]: ...



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from pathlib import Path
from unittest import TestCase

from libcst import parse_module
from libcst._parser.entrypoints import is_native

fixtures: Path = Path(__file__).parent.parent.parent / "native/libcst/tests/fixtures"


class RoundTripTests(TestCase):
    def test_clean_roundtrip(self) -> None:
        if not is_native():
            self.skipTest("pure python parser doesn't work with this")
        self.assertTrue(fixtures.exists(), f"{fixtures} should exist")
        files = list(fixtures.iterdir())
        self.assertGreater(len(files), 0)
        for file in files:
            with self.subTest(file=str(file)):
                src = file.read_text(encoding="utf-8")
                mod = parse_module(src)
                self.assertEqual(mod.code, src)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, FrozenSet, Mapping, Sequence

from libcst._parser.parso.utils import PythonVersionInfo

class BaseWhitespaceParserConfig:
    def __new__(
        cls,
        *,
        lines: Sequence[str],
        default_newline: str,
    ) -> BaseWhitespaceParserConfig: ...
    lines: Sequence[str]
    default_newline: str

class ParserConfig(BaseWhitespaceParserConfig):
    def __new__(
        cls,
        *,
        lines: Sequence[str],
        encoding: str,
        default_indent: str,
        default_newline: str,
        has_trailing_newline: bool,
        version: PythonVersionInfo,
        future_imports: FrozenSet[str],
    ) -> BaseWhitespaceParserConfig: ...
    # lines is inherited
    encoding: str
    default_indent: str
    # default_newline is inherited
    has_trailing_newline: bool
    version: PythonVersionInfo
    future_imports: FrozenSet[str]

def parser_config_asdict(config: ParserConfig) -> Mapping[str, Any]: ...



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
# pyre-unsafe

# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

project = "LibCST"
copyright = "Meta Platforms, Inc. and affiliates"
author = "Benjamin Woodruff, Jennifer Taylor, Carl Meyer, Jimmy Lai, Ray Zeng"

# The short X.Y version
version = ""
# The full version, including alpha/beta/rc tags
release = ""


# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "nbsphinx",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.graphviz",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
    "sphinx_rtd_theme",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "**.ipynb_checkpoints"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = None


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {"logo_only": True}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}

html_logo = "_static/logo/horizontal_white_sidebar.png"
html_favicon = "_static/logo/favicon.ico"

# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "LibCSTdoc"


# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [(master_doc, "LibCST.tex", "LibCST Documentation", author, "manual")]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "libcst", "LibCST Documentation", [author], 1)]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "LibCST",
        "LibCST Documentation",
        author,
        "LibCST",
        "One line description of project.",
        "Miscellaneous",
    )
]


# -- Options for Epub output -------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ["search.html"]


# -- Extension configuration -------------------------------------------------
autodoc_member_order = "bysource"
autodoc_default_options = {"members": True, "undoc-members": True}

intersphinx_mapping = {"python": ("https://docs.python.org/3", None)}

# -- Options for todo extension ----------------------------------------------

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

# -- autodoc customization
def strip_class_signature(app, what, name, obj, options, signature, return_annotation):
    if what == "class":
        return (None, return_annotation)
    return (signature, return_annotation)


def strip_class_signature_docstring(app, what, name, obj, options, lines):
    if what == "class":
        cls_name = name.split(".")[-1]
        if lines and lines[0].startswith(cls_name):
            while lines:
                del lines[0]


def setup(app):
    app.connect("autodoc-process-signature", strip_class_signature)
    app.connect("autodoc-process-docstring", strip_class_signature_docstring)
    app.add_css_file("custom.css")


nbsphinx_prolog = r"""
{% set docname = 'docs/source/' + env.doc2path(env.docname, base=None) %}

.. only:: html

    .. nbinfo::

        Interactive online tutorial: |notebook-badge|

.. |notebook-badge| image:: https://img.shields.io/badge/notebook-run-579ACA.svg?logo=data:image/png;base64,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
   :target: https://mybinder.org/v2/gh/Instagram/LibCST/main?filepath={{ docname }}
   :alt: Notebook
"""



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from unittest.mock import Mock

import libcst as cst
from libcst import parse_module
from libcst._exceptions import MetadataException
from libcst._visitors import CSTTransformer
from libcst.metadata import (
    BatchableMetadataProvider,
    MetadataWrapper,
    VisitorMetadataProvider,
)
from libcst.testing.utils import UnitTest


class MetadataProviderTest(UnitTest):
    def test_visitor_provider(self) -> None:
        """
        Tests that visitor providers are resolved correctly.

        Sets 2 metadata entries for every node:
            SimpleProvider -> 1
            DependentProvider - > 2
        """

        test = self

        class SimpleProvider(VisitorMetadataProvider[int]):
            def on_visit(self, node: cst.CSTNode) -> bool:
                self.set_metadata(node, 1)
                return True

        class DependentProvider(VisitorMetadataProvider[int]):
            METADATA_DEPENDENCIES = (SimpleProvider,)

            def on_visit(self, node: cst.CSTNode) -> bool:
                self.set_metadata(node, self.get_metadata(SimpleProvider, node) + 1)
                return True

        class DependentVisitor(CSTTransformer):
            # Declare both providers so the visitor has acesss to both types of metadata
            METADATA_DEPENDENCIES = (DependentProvider, SimpleProvider)

            def visit_Module(self, node: cst.Module) -> None:
                # Check metadata is set
                test.assertEqual(self.get_metadata(SimpleProvider, node), 1)
                test.assertEqual(self.get_metadata(DependentProvider, node), 2)

            def visit_Pass(self, node: cst.Pass) -> None:
                # Check metadata is set
                test.assertEqual(self.get_metadata(SimpleProvider, node), 1)
                test.assertEqual(self.get_metadata(DependentProvider, node), 2)

        module = parse_module("pass")
        MetadataWrapper(module).visit(DependentVisitor())

    def test_batched_provider(self) -> None:
        """
        Tests that batchable providers are resolved correctly.

        Sets metadata on:
            - pass: BatchedProviderA -> 1
                    BatchedProviderB -> "a"

        """
        test = self
        mock = Mock()

        class BatchedProviderA(BatchableMetadataProvider[int]):
            def visit_Pass(self, node: cst.Pass) -> None:
                mock.visited_a()
                self.set_metadata(node, 1)

        class BatchedProviderB(BatchableMetadataProvider[str]):
            def visit_Pass(self, node: cst.Pass) -> None:
                mock.visited_b()
                self.set_metadata(node, "a")

        class DependentVisitor(CSTTransformer):
            METADATA_DEPENDENCIES = (BatchedProviderA, BatchedProviderB)

            def visit_Pass(self, node: cst.Pass) -> None:
                # Check metadata is set
                test.assertEqual(self.get_metadata(BatchedProviderA, node), 1)
                test.assertEqual(self.get_metadata(BatchedProviderB, node), "a")

        module = parse_module("pass")
        MetadataWrapper(module).visit(DependentVisitor())

        # Check that each batchable visitor is only called once
        mock.visited_a.assert_called_once()
        mock.visited_b.assert_called_once()

    def test_mixed_providers(self) -> None:
        """
        Tests that a mixed set of providers is resolved properly.

        Sets metadata on pass:
            BatchedProviderA -> 2
            BatchedProviderB -> 3
            DependentProvider -> 5
            DependentBatched -> 4
        """
        test = self
        mock = Mock()

        class SimpleProvider(VisitorMetadataProvider[int]):
            def visit_Pass(self, node: cst.CSTNode) -> None:
                mock.visited_simple()
                self.set_metadata(node, 1)

        class BatchedProviderA(BatchableMetadataProvider[int]):
            METADATA_DEPENDENCIES = (SimpleProvider,)

            def visit_Pass(self, node: cst.Pass) -> None:
                mock.visited_a()
                self.set_metadata(node, 2)

        class BatchedProviderB(BatchableMetadataProvider[int]):
            METADATA_DEPENDENCIES = (SimpleProvider,)

            def visit_Pass(self, node: cst.Pass) -> None:
                mock.visited_b()
                self.set_metadata(node, 3)

        class DependentProvider(VisitorMetadataProvider[int]):
            METADATA_DEPENDENCIES = (BatchedProviderA, BatchedProviderB)

            def on_visit(self, node: cst.CSTNode) -> bool:
                sum = self.get_metadata(BatchedProviderA, node, 0) + self.get_metadata(
                    BatchedProviderB, node, 0
                )
                self.set_metadata(node, sum)
                return True

        class BatchedProviderC(BatchableMetadataProvider[int]):
            METADATA_DEPENDENCIES = (BatchedProviderA,)

            def visit_Pass(self, node: cst.Pass) -> None:
                mock.visited_c()
                self.set_metadata(node, self.get_metadata(BatchedProviderA, node) * 2)

        class DependentVisitor(CSTTransformer):
            METADATA_DEPENDENCIES = (
                BatchedProviderA,
                BatchedProviderB,
                BatchedProviderC,
                DependentProvider,
            )

            def visit_Module(self, node: cst.Module) -> None:
                # Dependent visitor set metadata on all nodes but for module it
                # defaulted to 0 because BatchedProviderA/B only set metadata on
                # pass nodes
                test.assertEqual(self.get_metadata(DependentProvider, node), 0)

            def visit_Pass(self, node: cst.Pass) -> None:
                # Check metadata is set
                test.assertEqual(self.get_metadata(BatchedProviderA, node), 2)
                test.assertEqual(self.get_metadata(BatchedProviderB, node), 3)
                test.assertEqual(self.get_metadata(BatchedProviderC, node), 4)
                test.assertEqual(self.get_metadata(DependentProvider, node), 5)

        module = parse_module("pass")
        MetadataWrapper(module).visit(DependentVisitor())

        # Check each visitor is called once
        mock.visited_simple.assert_called_once()
        mock.visited_a.assert_called_once()
        mock.visited_b.assert_called_once()
        mock.visited_c.assert_called_once()

    def test_inherited_metadata(self) -> None:
        """
        Tests that classes inherit access to metadata declared by their base
        classes.
        """
        test_runner = self
        mock = Mock()

        class SimpleProvider(VisitorMetadataProvider[int]):
            def visit_Pass(self, node: cst.Pass) -> None:
                mock.visited_simple()
                self.set_metadata(node, 1)

        class VisitorA(CSTTransformer):
            METADATA_DEPENDENCIES = (SimpleProvider,)

        class VisitorB(VisitorA):
            def visit_Pass(self, node: cst.Pass) -> None:
                test_runner.assertEqual(self.get_metadata(SimpleProvider, node), 1)

        module = parse_module("pass")
        MetadataWrapper(module).visit(VisitorB())

        # Check each visitor is called once
        mock.visited_simple.assert_called_once()

    def test_provider_inherited_metadata(self) -> None:
        """
        Tests that providers inherit access to metadata declared by their base
        classes.
        """
        test_runner = self
        mock = Mock()

        class ProviderA(VisitorMetadataProvider[int]):
            def visit_Pass(self, node: cst.Pass) -> None:
                mock.visited_a()
                self.set_metadata(node, 1)

        class ProviderB(VisitorMetadataProvider[int]):
            METADATA_DEPENDENCIES = (ProviderA,)

        class ProviderC(ProviderB):
            def visit_Pass(self, node: cst.Pass) -> None:
                mock.visited_c()
                test_runner.assertEqual(self.get_metadata(ProviderA, node), 1)

        class Visitor(CSTTransformer):
            METADATA_DEPENDENCIES = (ProviderC,)

        module = parse_module("pass")
        MetadataWrapper(module).visit(Visitor())

        # Check each visitor is called once
        mock.visited_a.assert_called_once()
        mock.visited_c.assert_called_once()

    def test_batchable_provider_inherited_metadata(self) -> None:
        """
        Tests that batchable providers inherit access to metadata declared by
        their base classes.
        """
        test_runner = self
        mock = Mock()

        class ProviderA(VisitorMetadataProvider[int]):
            def visit_Pass(self, node: cst.Pass) -> None:
                mock.visited_a()
                self.set_metadata(node, 1)

        class ProviderB(BatchableMetadataProvider[int]):
            METADATA_DEPENDENCIES = (ProviderA,)

        class ProviderC(ProviderB):
            def visit_Pass(self, node: cst.Pass) -> None:
                mock.visited_c()
                test_runner.assertEqual(self.get_metadata(ProviderA, node), 1)

        class VisitorA(CSTTransformer):
            METADATA_DEPENDENCIES = (ProviderC,)

        module = parse_module("pass")
        MetadataWrapper(module).visit(VisitorA())

        # Check each visitor is called once
        mock.visited_a.assert_called_once()
        mock.visited_c.assert_called_once()

    def test_self_metadata(self) -> None:
        """
        Tests a provider can access its own metadata (assuming it has been
        set properly.)
        """
        test_runner = self

        class ProviderA(VisitorMetadataProvider[bool]):
            def on_visit(self, node: cst.CSTNode) -> bool:
                self.set_metadata(node, True)
                return True

            def on_leave(self, original_node: cst.CSTNode) -> None:
                test_runner.assertEqual(
                    self.get_metadata(type(self), original_node), True
                )

        class AVisitor(CSTTransformer):
            METADATA_DEPENDENCIES = (ProviderA,)

        cst.Module([]).visit(AVisitor())

    def test_unset_metadata(self) -> None:
        """
        Tests that access to unset metadata throws a key error.
        """

        class ProviderA(VisitorMetadataProvider[bool]):
            pass

        class AVisitor(CSTTransformer):
            METADATA_DEPENDENCIES = (ProviderA,)

            def on_visit(self, node: cst.CSTNode) -> bool:
                self.get_metadata(ProviderA, node)
                return True

        with self.assertRaisesRegex(
            KeyError,
            "ProviderA is a dependency, but not set; did you forget a MetadataWrapper?",
        ):
            cst.Module([]).visit(AVisitor())

    def test_undeclared_metadata(self) -> None:
        """
        Tests that access to undeclared metadata throws a key error.
        """

        class ProviderA(VisitorMetadataProvider[bool]):
            pass

        class ProviderB(VisitorMetadataProvider[bool]):
            pass

        class AVisitor(CSTTransformer):
            METADATA_DEPENDENCIES = (ProviderA,)

            def on_visit(self, node: cst.CSTNode) -> bool:
                self.get_metadata(ProviderA, node, True)
                self.get_metadata(ProviderB, node)
                return True

        with self.assertRaisesRegex(
            KeyError,
            "ProviderB is not declared as a dependency in AVisitor.METADATA_DEPENDENCIES.",
        ):
            MetadataWrapper(cst.Module([])).visit(AVisitor())

    def test_circular_dependency(self) -> None:
        """
        Tests that circular dependencies are detected.
        """

        class ProviderA(VisitorMetadataProvider[str]):
            pass

        ProviderA.METADATA_DEPENDENCIES = (ProviderA,)

        class BadVisitor(CSTTransformer):
            METADATA_DEPENDENCIES = (ProviderA,)

        with self.assertRaisesRegex(
            MetadataException, "Detected circular dependencies in ProviderA"
        ):
            MetadataWrapper(cst.Module([])).visit(BadVisitor())



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import argparse
import textwrap
from typing import Dict, Optional

import libcst as cst
from libcst.codemod import CodemodContext, VisitorBasedCodemodCommand


presets_per_formatter: Dict[str, Dict[str, int]] = {
    "black": {
        "parameter_count": 1,
        "argument_count": 2,
    },
    "yapf": {
        "parameter_count": 2,
        "argument_count": 2,
    },
}


class AddTrailingCommas(VisitorBasedCodemodCommand):
    DESCRIPTION: str = textwrap.dedent(
        """
        Codemod that adds trailing commas to arguments in function
        headers and function calls.

        The idea is that both the black and yapf autoformatters will
        tend to split headers and function calls so that there
        is one parameter / argument per line if there is a trailing
        comma:
        - Black will always separate them by line
        - Yapf appears to do so whenever there are at least two arguments

        Applying this codemod (and then an autoformatter) may make
        it easier to read function definitions and calls
        """
    )

    def __init__(
        self,
        context: CodemodContext,
        formatter: str = "black",
        parameter_count: Optional[int] = None,
        argument_count: Optional[int] = None,
    ) -> None:
        super().__init__(context)
        presets = presets_per_formatter.get(formatter)
        if presets is None:
            raise ValueError(
                f"Unknown formatter {formatter!r}. Presets exist for "
                + ", ".join(presets_per_formatter.keys())
            )
        self.parameter_count: int = parameter_count or presets["parameter_count"]
        self.argument_count: int = argument_count or presets["argument_count"]

    @staticmethod
    def add_args(arg_parser: argparse.ArgumentParser) -> None:
        arg_parser.add_argument(
            "--formatter",
            dest="formatter",
            metavar="FORMATTER",
            help="Formatter to target (e.g. yapf or black)",
            type=str,
            default="black",
        )
        arg_parser.add_argument(
            "--paramter-count",
            dest="parameter_count",
            metavar="PARAMETER_COUNT",
            help="Minimal number of parameters for us to add trailing comma",
            type=int,
            default=None,
        )
        arg_parser.add_argument(
            "--argument-count",
            dest="argument_count",
            metavar="ARGUMENT_COUNT",
            help="Minimal number of arguments for us to add trailing comma",
            type=int,
            default=None,
        )

    def leave_Parameters(
        self,
        original_node: cst.Parameters,
        updated_node: cst.Parameters,
    ) -> cst.Parameters:
        skip = (
            #
            self.parameter_count is None
            or len(updated_node.params) < self.parameter_count
            or (
                len(updated_node.params) == 1
                and updated_node.params[0].name.value in {"self", "cls"}
            )
        )
        if skip:
            return updated_node
        else:
            last_param = updated_node.params[-1]
            return updated_node.with_changes(
                params=(
                    *updated_node.params[:-1],
                    last_param.with_changes(comma=cst.Comma()),
                ),
            )

    def leave_Call(
        self,
        original_node: cst.Call,
        updated_node: cst.Call,
    ) -> cst.Call:
        if len(updated_node.args) < self.argument_count:
            return updated_node
        else:
            last_arg = updated_node.args[-1]
            return updated_node.with_changes(
                args=(
                    *updated_node.args[:-1],
                    last_arg.with_changes(comma=cst.Comma()),
                ),
            )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Callable

import libcst as cst
from libcst._nodes.tests.base import CSTNodeTest
from libcst.testing.utils import data_provider


class NewlineTest(CSTNodeTest):
    @data_provider(
        (
            (cst.Newline("\r\n"), "\r\n"),
            (cst.Newline("\r"), "\r"),
            (cst.Newline("\n"), "\n"),
        )
    )
    def test_valid(self, node: cst.CSTNode, code: str) -> None:
        self.validate_node(node, code)

    @data_provider(
        (
            (lambda: cst.Newline("bad input"), "invalid value"),
            (lambda: cst.Newline("\nbad input\n"), "invalid value"),
            (lambda: cst.Newline("\n\n"), "invalid value"),
        )
    )
    def test_invalid(
        self, get_node: Callable[[], cst.CSTNode], expected_re: str
    ) -> None:
        self.assert_invalid(get_node, expected_re)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Set

import libcst
from libcst._visitors import CSTVisitor
from libcst.helpers.paths import chdir
from libcst.metadata import FilePathProvider, FullRepoManager, MetadataWrapper
from libcst.testing.utils import UnitTest


class FilePathProviderTest(UnitTest):
    def setUp(self) -> None:
        self.td = TemporaryDirectory()
        self.tdp = Path(self.td.name).resolve()
        self.addCleanup(self.td.cleanup)

    def test_provider_cache(self) -> None:
        pkg = self.tdp / "pkg"
        pkg.mkdir()
        files = [Path(pkg / name) for name in ("file1.py", "file2.py", "file3.py")]
        [file.write_text("print('hello')\n") for file in files]

        with self.subTest("absolute paths"):
            repo_manager = FullRepoManager(
                self.tdp, [f.as_posix() for f in files], {FilePathProvider}
            )
            repo_manager.resolve_cache()

            expected = {
                FilePathProvider: {f.as_posix(): f for f in files},
            }
            self.assertDictEqual(expected, repo_manager.cache)

        with self.subTest("repo relative paths"):
            repo_manager = FullRepoManager(
                self.tdp,
                [f.relative_to(self.tdp).as_posix() for f in files],
                {FilePathProvider},
            )
            repo_manager.resolve_cache()

            expected = {
                FilePathProvider: {
                    f.relative_to(self.tdp).as_posix(): f for f in files
                },
            }
            self.assertDictEqual(expected, repo_manager.cache)

        with self.subTest("dot relative paths"):
            with chdir(self.tdp):
                repo_manager = FullRepoManager(
                    ".",
                    [f.relative_to(self.tdp).as_posix() for f in files],
                    {FilePathProvider},
                )
                repo_manager.resolve_cache()

                expected = {
                    FilePathProvider: {
                        f.relative_to(self.tdp).as_posix(): f for f in files
                    },
                }
                self.assertDictEqual(expected, repo_manager.cache)

    def test_visitor(self) -> None:
        pkg = self.tdp / "pkg"
        pkg.mkdir()
        files = [Path(pkg / name) for name in ("file1.py", "file2.py", "file3.py")]
        [file.write_text("print('hello')\n") for file in files]

        seen: Set[Path] = set()

        class FakeVisitor(CSTVisitor):
            METADATA_DEPENDENCIES = [FilePathProvider]

            def visit_Module(self, node: libcst.Module) -> None:
                seen.add(self.get_metadata(FilePathProvider, node))

        with self.subTest("absolute paths"):
            seen.clear()
            repo_manager = FullRepoManager(
                self.tdp, [f.as_posix() for f in files], {FilePathProvider}
            )
            repo_manager.resolve_cache()

            for file in files:
                module = libcst.parse_module(file.read_bytes())
                wrapper = MetadataWrapper(
                    module, cache=repo_manager.get_cache_for_path(file.as_posix())
                )
                wrapper.visit(FakeVisitor())

            expected = set(files)
            self.assertSetEqual(expected, seen)

        with self.subTest("repo relative paths"):
            seen.clear()
            repo_manager = FullRepoManager(
                self.tdp,
                [f.relative_to(self.tdp).as_posix() for f in files],
                {FilePathProvider},
            )
            repo_manager.resolve_cache()

            for file in files:
                module = libcst.parse_module(file.read_bytes())
                wrapper = MetadataWrapper(
                    module,
                    cache=repo_manager.get_cache_for_path(
                        file.relative_to(self.tdp).as_posix()
                    ),
                )
                wrapper.visit(FakeVisitor())

            expected = set(files)
            self.assertSetEqual(expected, seen)

        with self.subTest("dot relative paths"):
            with chdir(self.tdp):
                seen.clear()
                repo_manager = FullRepoManager(
                    ".",
                    [f.relative_to(self.tdp).as_posix() for f in files],
                    {FilePathProvider},
                )
                repo_manager.resolve_cache()

                for file in files:
                    module = libcst.parse_module(file.read_bytes())
                    wrapper = MetadataWrapper(
                        module,
                        cache=repo_manager.get_cache_for_path(
                            file.relative_to(self.tdp).as_posix()
                        ),
                    )
                    wrapper.visit(FakeVisitor())

                expected = set(files)
                self.assertSetEqual(expected, seen)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
from typing import Tuple

import libcst as cst
import libcst.matchers as m
import libcst.metadata as meta
from libcst.testing.utils import UnitTest


class MatchersExtractTest(UnitTest):
    def _make_coderange(
        self, start: Tuple[int, int], end: Tuple[int, int]
    ) -> meta.CodeRange:
        return meta.CodeRange(
            start=meta.CodePosition(line=start[0], column=start[1]),
            end=meta.CodePosition(line=end[0], column=end[1]),
        )

    def test_extract_sentinel(self) -> None:
        # Verify behavior when provided a sentinel
        nothing = m.extract(
            cst.RemovalSentinel.REMOVE,
            m.Call(func=m.SaveMatchedNode(m.Name(), name="func")),
        )
        self.assertIsNone(nothing)
        nothing = m.extract(
            cst.MaybeSentinel.DEFAULT,
            m.Call(func=m.SaveMatchedNode(m.Name(), name="func")),
        )
        self.assertIsNone(nothing)

    def test_extract_tautology(self) -> None:
        expression = cst.parse_expression("a + b[c], d(e, f * g)")
        nodes = m.extract(
            expression,
            m.SaveMatchedNode(
                m.Tuple(elements=[m.Element(m.BinaryOperation()), m.Element(m.Call())]),
                name="node",
            ),
        )
        self.assertEqual(nodes, {"node": expression})

    def test_extract_simple(self) -> None:
        # Verify true behavior
        expression = cst.parse_expression("a + b[c], d(e, f * g)")
        nodes = m.extract(
            expression,
            m.Tuple(
                elements=[
                    m.Element(
                        m.BinaryOperation(left=m.SaveMatchedNode(m.Name(), "left"))
                    ),
                    m.Element(m.Call()),
                ]
            ),
        )
        extracted_node = cst.ensure_type(
            cst.ensure_type(expression, cst.Tuple).elements[0].value,
            cst.BinaryOperation,
        ).left
        self.assertEqual(nodes, {"left": extracted_node})

        # Verify false behavior
        nodes = m.extract(
            expression,
            m.Tuple(
                elements=[
                    m.Element(
                        m.BinaryOperation(left=m.SaveMatchedNode(m.Subscript(), "left"))
                    ),
                    m.Element(m.Call()),
                ]
            ),
        )
        self.assertIsNone(nodes)

    def test_extract_multiple(self) -> None:
        expression = cst.parse_expression("a + b[c], d(e, f * g)")
        nodes = m.extract(
            expression,
            m.Tuple(
                elements=[
                    m.Element(
                        m.BinaryOperation(left=m.SaveMatchedNode(m.Name(), "left"))
                    ),
                    m.Element(m.Call(func=m.SaveMatchedNode(m.Name(), "func"))),
                ]
            ),
        )
        extracted_node_left = cst.ensure_type(
            cst.ensure_type(expression, cst.Tuple).elements[0].value,
            cst.BinaryOperation,
        ).left
        extracted_node_func = cst.ensure_type(
            cst.ensure_type(expression, cst.Tuple).elements[1].value, cst.Call
        ).func
        self.assertEqual(
            nodes, {"left": extracted_node_left, "func": extracted_node_func}
        )

    def test_extract_predicates(self) -> None:
        expression = cst.parse_expression("a + b[c], d(e, f * g)")
        nodes = m.extract(
            expression,
            m.Tuple(
                elements=[
                    m.Element(
                        m.BinaryOperation(left=m.SaveMatchedNode(m.Name(), "left"))
                    ),
                    m.Element(
                        m.Call(
                            func=m.SaveMatchedNode(m.Name(), "func")
                            | m.SaveMatchedNode(m.Attribute(), "attr")
                        )
                    ),
                ]
            ),
        )
        extracted_node_left = cst.ensure_type(
            cst.ensure_type(expression, cst.Tuple).elements[0].value,
            cst.BinaryOperation,
        ).left
        extracted_node_func = cst.ensure_type(
            cst.ensure_type(expression, cst.Tuple).elements[1].value, cst.Call
        ).func
        self.assertEqual(
            nodes, {"left": extracted_node_left, "func": extracted_node_func}
        )

        expression = cst.parse_expression("a + b[c], d.z(e, f * g)")
        nodes = m.extract(
            expression,
            m.Tuple(
                elements=[
                    m.Element(
                        m.BinaryOperation(left=m.SaveMatchedNode(m.Name(), "left"))
                    ),
                    m.Element(
                        m.Call(
                            func=m.SaveMatchedNode(m.Name(), "func")
                            | m.SaveMatchedNode(m.Attribute(), "attr")
                        )
                    ),
                ]
            ),
        )
        extracted_node_left = cst.ensure_type(
            cst.ensure_type(expression, cst.Tuple).elements[0].value,
            cst.BinaryOperation,
        ).left
        extracted_node_attr = cst.ensure_type(
            cst.ensure_type(expression, cst.Tuple).elements[1].value, cst.Call
        ).func
        self.assertEqual(
            nodes, {"left": extracted_node_left, "attr": extracted_node_attr}
        )

    def test_extract_metadata(self) -> None:
        # Verify true behavior
        module = cst.parse_module("a + b[c], d(e, f * g)")
        wrapper = cst.MetadataWrapper(module)
        expression = cst.ensure_type(
            cst.ensure_type(wrapper.module.body[0], cst.SimpleStatementLine).body[0],
            cst.Expr,
        ).value

        nodes = m.extract(
            expression,
            m.Tuple(
                elements=[
                    m.Element(
                        m.BinaryOperation(
                            left=m.Name(
                                metadata=m.SaveMatchedNode(
                                    m.MatchMetadata(
                                        meta.PositionProvider,
                                        self._make_coderange((1, 0), (1, 1)),
                                    ),
                                    "left",
                                )
                            )
                        )
                    ),
                    m.Element(m.Call()),
                ]
            ),
            metadata_resolver=wrapper,
        )
        extracted_node = cst.ensure_type(
            cst.ensure_type(expression, cst.Tuple).elements[0].value,
            cst.BinaryOperation,
        ).left
        self.assertEqual(nodes, {"left": extracted_node})

        # Verify false behavior
        nodes = m.extract(
            expression,
            m.Tuple(
                elements=[
                    m.Element(
                        m.BinaryOperation(
                            left=m.Name(
                                metadata=m.SaveMatchedNode(
                                    m.MatchMetadata(
                                        meta.PositionProvider,
                                        self._make_coderange((1, 0), (1, 2)),
                                    ),
                                    "left",
                                )
                            )
                        )
                    ),
                    m.Element(m.Call()),
                ]
            ),
            metadata_resolver=wrapper,
        )
        self.assertIsNone(nodes)

    def test_extract_precedence_parent(self) -> None:
        expression = cst.parse_expression("a + b[c], d(e, f * g)")
        nodes = m.extract(
            expression,
            m.Tuple(
                elements=[
                    m.DoNotCare(),
                    m.Element(
                        m.SaveMatchedNode(
                            m.Call(
                                args=[
                                    m.Arg(m.SaveMatchedNode(m.Name(), "name")),
                                    m.DoNotCare(),
                                ]
                            ),
                            "name",
                        )
                    ),
                ]
            ),
        )
        extracted_node = cst.ensure_type(expression, cst.Tuple).elements[1].value
        self.assertEqual(nodes, {"name": extracted_node})

    def test_extract_precedence_sequence(self) -> None:
        expression = cst.parse_expression("a + b[c], d(e, f * g)")
        nodes = m.extract(
            expression,
            m.Tuple(
                elements=[
                    m.DoNotCare(),
                    m.Element(
                        m.Call(
                            args=[
                                m.Arg(m.SaveMatchedNode(m.DoNotCare(), "arg")),
                                m.Arg(m.SaveMatchedNode(m.DoNotCare(), "arg")),
                            ]
                        )
                    ),
                ]
            ),
        )
        extracted_node = (
            cst.ensure_type(
                cst.ensure_type(expression, cst.Tuple).elements[1].value, cst.Call
            )
            .args[1]
            .value
        )
        self.assertEqual(nodes, {"arg": extracted_node})

    def test_extract_precedence_sequence_wildcard(self) -> None:
        expression = cst.parse_expression("a + b[c], d(e, f * g)")
        nodes = m.extract(
            expression,
            m.Tuple(
                elements=[
                    m.DoNotCare(),
                    m.Element(
                        m.Call(
                            args=[
                                m.ZeroOrMore(
                                    m.Arg(m.SaveMatchedNode(m.DoNotCare(), "arg"))
                                )
                            ]
                        )
                    ),
                ]
            ),
        )
        extracted_node = (
            cst.ensure_type(
                cst.ensure_type(expression, cst.Tuple).elements[1].value, cst.Call
            )
            .args[1]
            .value
        )
        self.assertEqual(nodes, {"arg": extracted_node})

    def test_extract_optional_wildcard(self) -> None:
        expression = cst.parse_expression("a + b[c], d(e, f * g)")
        nodes = m.extract(
            expression,
            m.Tuple(
                elements=[
                    m.DoNotCare(),
                    m.Element(
                        m.Call(
                            args=[
                                m.ZeroOrMore(),
                                m.ZeroOrOne(
                                    m.Arg(m.SaveMatchedNode(m.Attribute(), "arg"))
                                ),
                            ]
                        )
                    ),
                ]
            ),
        )
        self.assertEqual(nodes, {})

    def test_extract_optional_wildcard_head(self) -> None:
        expression = cst.parse_expression("[3]")
        nodes = m.extract(
            expression,
            m.List(
                elements=[
                    m.SaveMatchedNode(m.ZeroOrMore(), "head1"),
                    m.SaveMatchedNode(m.ZeroOrMore(), "head2"),
                    m.Element(value=m.Integer(value="3")),
                ]
            ),
        )
        self.assertEqual(nodes, {"head1": (), "head2": ()})

    def test_extract_optional_wildcard_tail(self) -> None:
        expression = cst.parse_expression("[3]")
        nodes = m.extract(
            expression,
            m.List(
                elements=[
                    m.Element(value=m.Integer(value="3")),
                    m.SaveMatchedNode(m.ZeroOrMore(), "tail1"),
                    m.SaveMatchedNode(m.ZeroOrMore(), "tail2"),
                ]
            ),
        )
        self.assertEqual(nodes, {"tail1": (), "tail2": ()})

    def test_extract_optional_wildcard_present(self) -> None:
        expression = cst.parse_expression("a + b[c], d(e, f * g, h.i.j)")
        nodes = m.extract(
            expression,
            m.Tuple(
                elements=[
                    m.DoNotCare(),
                    m.Element(
                        m.Call(
                            args=[
                                m.DoNotCare(),
                                m.DoNotCare(),
                                m.ZeroOrOne(
                                    m.Arg(m.SaveMatchedNode(m.Attribute(), "arg"))
                                ),
                            ]
                        )
                    ),
                ]
            ),
        )
        extracted_node = (
            cst.ensure_type(
                cst.ensure_type(expression, cst.Tuple).elements[1].value, cst.Call
            )
            .args[2]
            .value
        )
        self.assertEqual(nodes, {"arg": extracted_node})

    def test_extract_sequence(self) -> None:
        expression = cst.parse_expression("a + b[c], d(e, f * g, h.i.j)")
        nodes = m.extract(
            expression,
            m.Tuple(
                elements=[
                    m.DoNotCare(),
                    m.Element(m.Call(args=m.SaveMatchedNode([m.ZeroOrMore()], "args"))),
                ]
            ),
        )
        extracted_seq = cst.ensure_type(
            cst.ensure_type(expression, cst.Tuple).elements[1].value, cst.Call
        ).args
        self.assertEqual(nodes, {"args": extracted_seq})

    def test_extract_sequence_element(self) -> None:
        # Verify true behavior
        expression = cst.parse_expression("a + b[c], d(e, f * g, h.i.j)")
        nodes = m.extract(
            expression,
            m.Tuple(
                elements=[
                    m.DoNotCare(),
                    m.Element(m.Call(args=[m.SaveMatchedNode(m.ZeroOrMore(), "args")])),
                ]
            ),
        )
        extracted_seq = tuple(
            cst.ensure_type(
                cst.ensure_type(expression, cst.Tuple).elements[1].value, cst.Call
            ).args
        )
        self.assertEqual(nodes, {"args": extracted_seq})

        # Verify false behavior
        nodes = m.extract(
            expression,
            m.Tuple(
                elements=[
                    m.DoNotCare(),
                    m.Element(
                        m.Call(
                            args=[
                                m.SaveMatchedNode(
                                    m.ZeroOrMore(m.Arg(m.Subscript())), "args"
                                )
                            ]
                        )
                    ),
                ]
            ),
        )
        self.assertIsNone(nodes)

    def test_extract_sequence_multiple_wildcards(self) -> None:
        expression = cst.parse_expression("1, 2, 3, 4")
        nodes = m.extract(
            expression,
            m.Tuple(
                elements=(
                    m.SaveMatchedNode(m.ZeroOrMore(), "head"),
                    m.SaveMatchedNode(m.Element(value=m.Integer(value="3")), "element"),
                    m.SaveMatchedNode(m.ZeroOrMore(), "tail"),
                )
            ),
        )
        tuple_elements = cst.ensure_type(expression, cst.Tuple).elements
        self.assertEqual(
            nodes,
            {
                "head": tuple(tuple_elements[:2]),
                "element": tuple_elements[2],
                "tail": tuple(tuple_elements[3:]),
            },
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from libcst._parser.grammar import _should_include
from libcst._parser.parso.utils import PythonVersionInfo
from libcst.testing.utils import data_provider, UnitTest


class VersionCompareTest(UnitTest):
    @data_provider(
        (
            # Simple equality
            ("==3.6", PythonVersionInfo(3, 6), True),
            ("!=3.6", PythonVersionInfo(3, 6), False),
            # Equal or GT/LT
            (">=3.6", PythonVersionInfo(3, 5), False),
            (">=3.6", PythonVersionInfo(3, 6), True),
            (">=3.6", PythonVersionInfo(3, 7), True),
            ("<=3.6", PythonVersionInfo(3, 5), True),
            ("<=3.6", PythonVersionInfo(3, 6), True),
            ("<=3.6", PythonVersionInfo(3, 7), False),
            # GT/LT
            (">3.6", PythonVersionInfo(3, 5), False),
            (">3.6", PythonVersionInfo(3, 6), False),
            (">3.6", PythonVersionInfo(3, 7), True),
            ("<3.6", PythonVersionInfo(3, 5), True),
            ("<3.6", PythonVersionInfo(3, 6), False),
            ("<3.6", PythonVersionInfo(3, 7), False),
            # Multiple checks
            (">3.6,<3.8", PythonVersionInfo(3, 6), False),
            (">3.6,<3.8", PythonVersionInfo(3, 7), True),
            (">3.6,<3.8", PythonVersionInfo(3, 8), False),
        )
    )
    def test_tokenize(
        self,
        requested_version: str,
        actual_version: PythonVersionInfo,
        expected_result: bool,
    ) -> None:
        self.assertEqual(
            _should_include(requested_version, actual_version), expected_result
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import inspect
from typing import (
    Callable,
    cast,
    Iterable,
    List,
    Mapping,
    MutableMapping,
    Optional,
    TYPE_CHECKING,
)

from libcst._metadata_dependent import MetadataDependent
from libcst._typed_visitor import CSTTypedVisitorFunctions
from libcst._visitors import CSTNodeT, CSTVisitor

if TYPE_CHECKING:
    from libcst._nodes.base import CSTNode  # noqa: F401

VisitorMethod = Callable[["CSTNode"], None]
_VisitorMethodCollection = Mapping[str, List[VisitorMethod]]


class BatchableCSTVisitor(CSTTypedVisitorFunctions, MetadataDependent):
    """
    The low-level base visitor class for traversing a CST as part of a batched
    set of traversals. This should be used in conjunction with the
    :func:`~libcst.visit_batched` function or the
    :func:`~libcst.MetadataWrapper.visit_batched` method from
    :class:`~libcst.MetadataWrapper` to visit a tree.
    Instances of this class cannot modify the tree.
    """

    def get_visitors(self) -> Mapping[str, VisitorMethod]:
        """
        Returns a mapping of all the ``visit_<Type[CSTNode]>``,
        ``visit_<Type[CSTNode]>_<attribute>``, ``leave_<Type[CSTNode]>`` and
        `leave_<Type[CSTNode]>_<attribute>`` methods defined by this visitor,
        excluding all empty stubs.
        """

        methods = inspect.getmembers(
            self,
            lambda m: (
                inspect.ismethod(m)
                and (m.__name__.startswith("visit_") or m.__name__.startswith("leave_"))
                and not getattr(m, "_is_no_op", False)
            ),
        )

        # TODO: verify all visitor methods reference valid node classes.
        # for name, __ in methods:
        #     ...

        return dict(methods)


def visit_batched(
    node: CSTNodeT,
    batchable_visitors: Iterable[BatchableCSTVisitor],
    before_visit: Optional[VisitorMethod] = None,
    after_leave: Optional[VisitorMethod] = None,
) -> CSTNodeT:
    """
    Do a batched traversal over ``node`` with all ``visitors``.

    ``before_visit`` and ``after_leave`` are provided as optional hooks to
    execute before the ``visit_<Type[CSTNode]>`` and after the
    ``leave_<Type[CSTNode]>`` methods from each visitor in ``visitor`` are
    executed by the batched visitor.

    This function does not handle metadata dependency resolution for ``visitors``.
    See :func:`~libcst.MetadataWrapper.visit_batched` from
    :class:`~libcst.MetadataWrapper` for batched traversal with metadata dependency
    resolution.
    """
    visitor_methods = _get_visitor_methods(batchable_visitors)
    batched_visitor = _BatchedCSTVisitor(
        visitor_methods, before_visit=before_visit, after_leave=after_leave
    )
    return cast(CSTNodeT, node.visit(batched_visitor))


def _get_visitor_methods(
    batchable_visitors: Iterable[BatchableCSTVisitor],
) -> _VisitorMethodCollection:
    """
    Gather all ``visit_<Type[CSTNode]>``, ``visit_<Type[CSTNode]>_<attribute>``,
    ``leave_<Type[CSTNode]>`` amd `leave_<Type[CSTNode]>_<attribute>`` methods
    from ``batchabled_visitors``.
    """
    visitor_methods: MutableMapping[str, List[VisitorMethod]] = {}
    for bv in batchable_visitors:
        for name, fn in bv.get_visitors().items():
            visitor_methods.setdefault(name, []).append(fn)
    return visitor_methods


class _BatchedCSTVisitor(CSTVisitor):
    """
    Internal visitor class to perform batched traversal over a tree.
    """

    visitor_methods: _VisitorMethodCollection
    before_visit: Optional[VisitorMethod]
    after_leave: Optional[VisitorMethod]

    def __init__(
        self,
        visitor_methods: _VisitorMethodCollection,
        *,
        before_visit: Optional[VisitorMethod] = None,
        after_leave: Optional[VisitorMethod] = None,
    ) -> None:
        super().__init__()
        self.visitor_methods = visitor_methods
        self.before_visit = before_visit
        self.after_leave = after_leave

    def on_visit(self, node: "CSTNode") -> bool:
        """
        Call appropriate visit methods on node before visiting children.
        """
        before_visit = self.before_visit
        if before_visit is not None:
            before_visit(node)
        type_name = type(node).__name__
        for v in self.visitor_methods.get(f"visit_{type_name}", []):
            v(node)
        return True

    def on_leave(self, original_node: "CSTNode") -> None:
        """
        Call appropriate leave methods on node after visiting children.
        """
        type_name = type(original_node).__name__
        for v in self.visitor_methods.get(f"leave_{type_name}", []):
            v(original_node)
        after_leave = self.after_leave
        if after_leave is not None:
            after_leave(original_node)

    def on_visit_attribute(self, node: "CSTNode", attribute: str) -> None:
        """
        Call appropriate visit attribute methods on node before visiting
        attribute's children.
        """
        type_name = type(node).__name__
        for v in self.visitor_methods.get(f"visit_{type_name}_{attribute}", []):
            v(node)

    def on_leave_attribute(self, original_node: "CSTNode", attribute: str) -> None:
        """
        Call appropriate leave attribute methods on node after visiting
        attribute's children.
        """
        type_name = type(original_node).__name__
        for v in self.visitor_methods.get(f"leave_{type_name}_{attribute}", []):
            v(original_node)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
import libcst as cst
import libcst.matchers as m
from libcst.codemod import CodemodContext, CodemodTest, VisitorBasedCodemodCommand
from libcst.codemod.visitors import AddImportsVisitor, RemoveImportsVisitor
from libcst.metadata import (
    QualifiedName,
    QualifiedNameProvider,
    QualifiedNameSource,
    ScopeProvider,
)
from libcst.testing.utils import data_provider


class TestRemoveImportsCodemod(CodemodTest):
    TRANSFORM = RemoveImportsVisitor

    def test_noop(self) -> None:
        """
        Should do nothing.
        """

        before = """
            def foo() -> None:
                pass
        """
        after = """
            def foo() -> None:
                pass
        """

        self.assertCodemod(before, after, [])

    def test_remove_import_simple(self) -> None:
        """
        Should remove module as import
        """

        before = """
            import bar
            import baz

            def foo() -> None:
                pass
        """
        after = """
            import bar

            def foo() -> None:
                pass
        """

        self.assertCodemod(before, after, [("baz", None, None)])

    def test_remove_fromimport_simple(self) -> None:
        before = "from a import b, c"
        after = "from a import c"
        self.assertCodemod(before, after, [("a", "b", None)])

    def test_remove_fromimport_keeping_standalone_comment(self) -> None:
        before = """
            from foo import (
                bar,
                # comment
                baz,
            )
            from loooong import (
                bar,
                # comment
                short,
                this_stays
            )
            from third import (
                # comment
                short,
                this_stays_too
            )
            from fourth import (
                a,
                # comment
                b,
                c
            )
        """
        after = """
            from foo import (
                # comment
                baz,
            )
            from loooong import (
                this_stays
            )
            from third import (
                this_stays_too
            )
            from fourth import (
                a,
                c
            )
        """
        self.assertCodemod(
            before,
            after,
            [
                ("foo", "bar", None),
                ("loooong", "short", None),
                ("loooong", "bar", None),
                ("third", "short", None),
                ("fourth", "b", None),
            ],
        )

    def test_remove_fromimport_keeping_inline_comment(self) -> None:
        before = """
            from foo import (  # comment
                bar,
                # comment2
                baz,
            )
            from loooong import (
                bar,
                short,  # comment
                # comment2
                this_stays
            )
            from third import (
                short,  # comment
                this_stays_too  # comment2
            )
        """
        after = """
            from foo import (  # comment
                # comment2
                baz,
            )
            from loooong import (
                # comment2
                this_stays
            )
            from third import (
                this_stays_too  # comment2
            )
        """
        self.assertCodemod(
            before,
            after,
            [
                ("foo", "bar", None),
                ("loooong", "short", None),
                ("loooong", "bar", None),
                ("third", "short", None),
            ],
        )

    def test_remove_import_alias_simple(self) -> None:
        """
        Should remove aliased module as import
        """

        before = """
            import bar
            import baz as qux

            def foo() -> None:
                pass
        """
        after = """
            import bar

            def foo() -> None:
                pass
        """

        self.assertCodemod(before, after, [("baz", None, "qux")])

    def test_dont_remove_import_simple(self) -> None:
        """
        Should not remove module import with reference
        """

        before = """
            import bar
            import baz

            def foo() -> None:
                baz.qux()
        """
        after = """
            import bar
            import baz

            def foo() -> None:
                baz.qux()
        """

        self.assertCodemod(before, after, [("baz", None, None)])

    def test_dont_remove_import_alias_simple(self) -> None:
        """
        Should not remove aliased module import with reference
        """

        before = """
            import bar
            import baz as qux

            def foo() -> None:
                qux.quux()
        """
        after = """
            import bar
            import baz as qux

            def foo() -> None:
                qux.quux()
        """

        self.assertCodemod(before, after, [("baz", None, "qux")])

    def test_dont_remove_import_simple_wrong_alias(self) -> None:
        """
        Should not remove module as import since wrong alias
        """

        before = """
            import bar
            import baz

            def foo() -> None:
                pass
        """
        after = """
            import bar
            import baz

            def foo() -> None:
                pass
        """

        self.assertCodemod(before, after, [("baz", None, "qux")])

    def test_dont_remove_import_wrong_alias_simple(self) -> None:
        """
        Should not remove wrong aliased module as import
        """

        before = """
            import bar
            import baz as qux

            def foo() -> None:
                pass
        """
        after = """
            import bar
            import baz as qux

            def foo() -> None:
                pass
        """

        self.assertCodemod(before, after, [("baz", None, None)])

    def test_remove_importfrom_simple(self) -> None:
        """
        Should remove import from
        """

        before = """
            import bar
            from baz import qux

            def foo() -> None:
                pass
        """
        after = """
            import bar

            def foo() -> None:
                pass
        """

        self.assertCodemod(before, after, [("baz", "qux", None)])

    def test_remove_importfrom_alias_simple(self) -> None:
        """
        Should remove import from with alias
        """

        before = """
            import bar
            from baz import qux as quux

            def foo() -> None:
                pass
        """
        after = """
            import bar

            def foo() -> None:
                pass
        """

        self.assertCodemod(before, after, [("baz", "qux", "quux")])

    def test_dont_remove_importfrom_simple(self) -> None:
        """
        Should not remove import from with reference
        """

        before = """
            import bar
            from baz import qux

            def foo() -> None:
                qux()
        """
        after = """
            import bar
            from baz import qux

            def foo() -> None:
                qux()
        """

        self.assertCodemod(before, after, [("baz", "qux", None)])

    def test_dont_remove_importfrom_alias_simple(self) -> None:
        """
        Should not remove aliased import from with reference
        """

        before = """
            import bar
            from baz import qux as quux

            def foo() -> None:
                quux()
        """
        after = """
            import bar
            from baz import qux as quux

            def foo() -> None:
                quux()
        """

        self.assertCodemod(before, after, [("baz", "qux", "quux")])

    def test_dont_remove_importfrom_simple_wrong_alias(self) -> None:
        """
        Should not remove import from since it is wrong alias
        """

        before = """
            import bar
            from baz import qux as quux

            def foo() -> None:
                pass
        """
        after = """
            import bar
            from baz import qux as quux

            def foo() -> None:
                pass
        """

        self.assertCodemod(before, after, [("baz", "qux", None)])

    def test_dont_remove_importfrom_alias_simple_wrong_alias(self) -> None:
        """
        Should not remove import from with wrong alias
        """

        before = """
            import bar
            from baz import qux

            def foo() -> None:
                pass
        """
        after = """
            import bar
            from baz import qux

            def foo() -> None:
                pass
        """

        self.assertCodemod(before, after, [("baz", "qux", "quux")])

    def test_remove_importfrom_relative(self) -> None:
        """
        Should remove import from which is relative
        """

        before = """
            import bar
            from .c import qux

            def foo() -> None:
                pass
        """
        after = """
            import bar

            def foo() -> None:
                pass
        """

        self.assertCodemod(
            before,
            after,
            [("a.b.c", "qux", None)],
            context_override=CodemodContext(
                full_module_name="a.b.foobar", full_package_name="a.b"
            ),
        )

    def test_dont_remove_inuse_importfrom_relative(self) -> None:
        """
        Should not remove import from which is relative since it is in use.
        """

        before = """
            import bar
            from .c import qux

            def foo() -> None:
                qux()
        """
        after = """
            import bar
            from .c import qux

            def foo() -> None:
                qux()
        """

        self.assertCodemod(
            before,
            after,
            [("a.b.c", "qux", None)],
            context_override=CodemodContext(
                full_module_name="a.b.foobar", full_package_name="a.b"
            ),
        )

    def test_dont_remove_wrong_importfrom_relative(self) -> None:
        """
        Should not remove import from which is relative since it is the wrong module.
        """

        before = """
            import bar
            from .c import qux

            def foo() -> None:
                pass
        """
        after = """
            import bar
            from .c import qux

            def foo() -> None:
                pass
        """

        self.assertCodemod(
            before,
            after,
            [("a.b.d", "qux", None)],
            context_override=CodemodContext(
                full_module_name="a.b.foobar", full_package_name="a.b"
            ),
        )

    def test_remove_import_complex(self) -> None:
        """
        Should remove complex module as import
        """

        before = """
            import bar
            import baz, qux
            import a.b
            import c.d
            import x.y.z
            import e.f as g
            import h.i as j

            def foo() -> None:
                c.d()
                x.u
                j()
        """
        after = """
            import bar
            import qux
            import c.d
            import x.y.z
            import h.i as j

            def foo() -> None:
                c.d()
                x.u
                j()
        """

        self.assertCodemod(
            before,
            after,
            [
                ("baz", None, None),
                ("a.b", None, None),
                ("c.d", None, None),
                ("e.f", None, "g"),
                ("h.i", None, "j"),
                ("x.y.z", None, None),
            ],
        )

    def test_remove_fromimport_complex(self) -> None:
        """
        Should remove complex from import
        """

        before = """
            from bar import qux, quux
            from a.b import c
            from d.e import f
            from h.i import j as k
            from l.m import n as o
            from x import *

            def foo() -> None:
                f()
                k()
        """
        after = """
            from bar import qux
            from d.e import f
            from h.i import j as k
            from x import *

            def foo() -> None:
                f()
                k()
        """

        self.assertCodemod(
            before,
            after,
            [
                ("bar", "quux", None),
                ("a.b", "c", None),
                ("d.e", "f", None),
                ("h.i", "j", "k"),
                ("l.m", "n", "o"),
            ],
        )

    def test_remove_import_multiple_assignments(self) -> None:
        """
        Should not remove import with multiple assignments
        """

        before = """
            from foo import bar
            from qux import bar

            def foo() -> None:
                bar()
        """
        after = """
            from foo import bar
            from qux import bar

            def foo() -> None:
                bar()
        """

        self.assertCodemod(before, after, [("foo", "bar", None)])

    def test_remove_multiple_imports(self) -> None:
        """
        Multiple imports
        """
        before = """
            try:
                import a
            except Exception:
                import a

            a.hello()
        """
        after = """
            try:
                import a
            except Exception:
                import a

            a.hello()
        """
        self.assertCodemod(before, after, [("a", None, None)])

        before = """
            try:
                import a
            except Exception:
                import a
        """
        after = """
            try:
                pass
            except Exception:
                pass
        """
        self.assertCodemod(before, after, [("a", None, None)])

    @data_provider(
        (
            # Simple removal, no other uses.
            (
                """
                    from foo import bar
                    from qux import baz

                    def fun() -> None:
                        bar()
                        baz()
                """,
                """
                    from qux import baz

                    def fun() -> None:
                        baz()
                """,
            ),
            # Remove a node, other uses, don't remove import.
            (
                """
                    from foo import bar
                    from qux import baz

                    def fun() -> None:
                        bar()
                        baz()

                    def foobar() -> None:
                        a = bar
                        a()
                """,
                """
                    from foo import bar
                    from qux import baz

                    def fun() -> None:
                        baz()

                    def foobar() -> None:
                        a = bar
                        a()
                """,
            ),
            # Remove an alias.
            (
                """
                    from foo import bar as other
                    from qux import baz

                    def fun() -> None:
                        other()
                        baz()
                """,
                """
                    from qux import baz

                    def fun() -> None:
                        baz()
                """,
            ),
            # Simple removal, no other uses.
            (
                """
                    import foo
                    from qux import baz

                    def fun() -> None:
                        foo.bar()
                        baz()
                """,
                """
                    from qux import baz

                    def fun() -> None:
                        baz()
                """,
            ),
            # Remove a node, other uses, don't remove import.
            (
                """
                    import foo
                    from qux import baz

                    def fun() -> None:
                        foo.bar()
                        baz()

                    def foobar() -> None:
                        a = foo.bar
                        a()
                """,
                """
                    import foo
                    from qux import baz

                    def fun() -> None:
                        baz()

                    def foobar() -> None:
                        a = foo.bar
                        a()
                """,
            ),
            # Remove an alias.
            (
                """
                    import foo as other
                    from qux import baz

                    def fun() -> None:
                        other.bar()
                        baz()
                """,
                """
                    from qux import baz

                    def fun() -> None:
                        baz()
                """,
            ),
        )
    )
    def test_remove_import_by_node_simple(self, before: str, after: str) -> None:
        """
        Given a node that's directly referenced in an import,
        make sure that the import is removed when the node
        is also removed.
        """

        class RemoveBarTransformer(VisitorBasedCodemodCommand):
            METADATA_DEPENDENCIES = (QualifiedNameProvider, ScopeProvider)

            @m.leave(
                m.SimpleStatementLine(
                    body=[
                        m.Expr(
                            m.Call(
                                metadata=m.MatchMetadata(
                                    QualifiedNameProvider,
                                    {
                                        QualifiedName(
                                            source=QualifiedNameSource.IMPORT,
                                            name="foo.bar",
                                        )
                                    },
                                )
                            )
                        )
                    ]
                )
            )
            def _leave_foo_bar(
                self,
                original_node: cst.SimpleStatementLine,
                updated_node: cst.SimpleStatementLine,
            ) -> cst.RemovalSentinel:
                RemoveImportsVisitor.remove_unused_import_by_node(
                    self.context, original_node
                )
                return cst.RemoveFromParent()

        module = cst.parse_module(self.make_fixture_data(before))
        self.assertCodeEqual(
            after, RemoveBarTransformer(CodemodContext()).transform_module(module).code
        )

    def test_remove_import_from_node(self) -> None:
        """
        Make sure that if an import node itself is requested for
        removal, we still do the right thing and only remove it
        if it is unused.
        """

        before = """
            from foo import bar
            from qux import baz
            from foo import qux as other
            from qux import foobar as other2

            def fun() -> None:
                baz()
                other2()
        """
        after = """
            from qux import baz
            from qux import foobar as other2

            def fun() -> None:
                baz()
                other2()
        """

        class RemoveImportTransformer(VisitorBasedCodemodCommand):
            METADATA_DEPENDENCIES = (QualifiedNameProvider, ScopeProvider)

            def visit_ImportFrom(self, node: cst.ImportFrom) -> None:
                RemoveImportsVisitor.remove_unused_import_by_node(self.context, node)

        module = cst.parse_module(self.make_fixture_data(before))
        self.assertCodeEqual(
            after,
            RemoveImportTransformer(CodemodContext()).transform_module(module).code,
        )

    def test_remove_import_node(self) -> None:
        """
        Make sure that if an import node itself is requested for
        removal, we still do the right thing and only remove it
        if it is unused.
        """

        before = """
            import foo
            import qux
            import bar as other
            import foobar as other2

            def fun() -> None:
                qux.baz()
                other2.baz()
        """
        after = """
            import qux
            import foobar as other2

            def fun() -> None:
                qux.baz()
                other2.baz()
        """

        class RemoveImportTransformer(VisitorBasedCodemodCommand):
            METADATA_DEPENDENCIES = (QualifiedNameProvider, ScopeProvider)

            def visit_Import(self, node: cst.Import) -> None:
                RemoveImportsVisitor.remove_unused_import_by_node(self.context, node)

        module = cst.parse_module(self.make_fixture_data(before))
        self.assertCodeEqual(
            after,
            RemoveImportTransformer(CodemodContext()).transform_module(module).code,
        )

    def test_remove_import_with_all(self) -> None:
        """
        Make sure that if an import node itself is requested for
        removal, we don't remove it if it shows up in an __all__
        node.
        """

        before = """
            from foo import bar
            from qux import baz

            __all__ = ["baz"]
        """
        after = """
            from qux import baz

            __all__ = ["baz"]
        """

        class RemoveImportTransformer(VisitorBasedCodemodCommand):
            METADATA_DEPENDENCIES = (QualifiedNameProvider, ScopeProvider)

            def visit_ImportFrom(self, node: cst.ImportFrom) -> None:
                RemoveImportsVisitor.remove_unused_import_by_node(self.context, node)

        module = cst.parse_module(self.make_fixture_data(before))
        self.assertCodeEqual(
            after,
            RemoveImportTransformer(CodemodContext()).transform_module(module).code,
        )

    def test_remove_import_alias_after_inserting(self) -> None:
        before = "from foo import bar, baz"
        after = "from foo import quux, baz"

        class AddRemoveTransformer(VisitorBasedCodemodCommand):
            def visit_Module(self, node: cst.Module) -> None:
                AddImportsVisitor.add_needed_import(self.context, "foo", "quux")
                RemoveImportsVisitor.remove_unused_import(self.context, "foo", "bar")

        module = cst.parse_module(self.make_fixture_data(before))
        self.assertCodeEqual(
            AddRemoveTransformer(CodemodContext()).transform_module(module).code,
            after,
        )

    def test_remove_comma(self) -> None:
        """
        Trailing commas should be removed if and only if the last alias is removed.
        """
        before = """
            from m import (a, b,)
            import x, y
        """
        after = """
            from m import (b,)
            import x
        """
        self.assertCodemod(before, after, [("m", "a", None), ("y", None, None)])



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


"""
Data structures used for storing position information.

These are publicly exported by metadata, but their implementation lives outside of
metadata, because they're used internally by the codegen logic, which computes position
locations.
"""

from dataclasses import dataclass
from typing import cast, overload, Tuple, Union

from libcst._add_slots import add_slots

_CodePositionT = Union[Tuple[int, int], "CodePosition"]


@add_slots
@dataclass(frozen=True)
class CodePosition:
    #: Line numbers are 1-indexed.
    line: int
    #: Column numbers are 0-indexed.
    column: int


@add_slots
@dataclass(frozen=True)
# pyre-fixme[13]: Attribute `end` is never initialized.
# pyre-fixme[13]: Attribute `start` is never initialized.
class CodeRange:
    #: Starting position of a node (inclusive).
    start: CodePosition
    #: Ending position of a node (exclusive).
    end: CodePosition

    @overload
    def __init__(self, start: CodePosition, end: CodePosition) -> None:
        ...

    @overload
    def __init__(self, start: Tuple[int, int], end: Tuple[int, int]) -> None:
        ...

    def __init__(self, start: _CodePositionT, end: _CodePositionT) -> None:
        if isinstance(start, tuple) and isinstance(end, tuple):
            object.__setattr__(self, "start", CodePosition(start[0], start[1]))
            object.__setattr__(self, "end", CodePosition(end[0], end[1]))
        else:
            start = cast(CodePosition, start)
            end = cast(CodePosition, end)
            object.__setattr__(self, "start", start)
            object.__setattr__(self, "end", end)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Callable, Optional

import libcst as cst
from libcst import parse_expression
from libcst._nodes.tests.base import CSTNodeTest
from libcst.metadata import CodeRange
from libcst.testing.utils import data_provider


class NumberTest(CSTNodeTest):
    @data_provider(
        (
            # Simple number
            (cst.Integer("5"), "5", parse_expression),
            # Negted number
            (
                cst.UnaryOperation(operator=cst.Minus(), expression=cst.Integer("5")),
                "-5",
                parse_expression,
                CodeRange((1, 0), (1, 2)),
            ),
            # In parenthesis
            (
                cst.UnaryOperation(
                    lpar=(cst.LeftParen(),),
                    operator=cst.Minus(),
                    expression=cst.Integer("5"),
                    rpar=(cst.RightParen(),),
                ),
                "(-5)",
                parse_expression,
                CodeRange((1, 1), (1, 3)),
            ),
            (
                cst.UnaryOperation(
                    lpar=(cst.LeftParen(),),
                    operator=cst.Minus(),
                    expression=cst.Integer(
                        "5", lpar=(cst.LeftParen(),), rpar=(cst.RightParen(),)
                    ),
                    rpar=(cst.RightParen(),),
                ),
                "(-(5))",
                parse_expression,
                CodeRange((1, 1), (1, 5)),
            ),
            (
                cst.UnaryOperation(
                    operator=cst.Minus(),
                    expression=cst.UnaryOperation(
                        operator=cst.Minus(), expression=cst.Integer("5")
                    ),
                ),
                "--5",
                parse_expression,
                CodeRange((1, 0), (1, 3)),
            ),
            # multiple nested parenthesis
            (
                cst.Integer(
                    "5",
                    lpar=(cst.LeftParen(), cst.LeftParen()),
                    rpar=(cst.RightParen(), cst.RightParen()),
                ),
                "((5))",
                parse_expression,
                CodeRange((1, 2), (1, 3)),
            ),
            (
                cst.UnaryOperation(
                    lpar=(cst.LeftParen(),),
                    operator=cst.Plus(),
                    expression=cst.Integer(
                        "5",
                        lpar=(cst.LeftParen(), cst.LeftParen()),
                        rpar=(cst.RightParen(), cst.RightParen()),
                    ),
                    rpar=(cst.RightParen(),),
                ),
                "(+((5)))",
                parse_expression,
                CodeRange((1, 1), (1, 7)),
            ),
        )
    )
    def test_valid(
        self,
        node: cst.CSTNode,
        code: str,
        parser: Optional[Callable[[str], cst.CSTNode]],
        position: Optional[CodeRange] = None,
    ) -> None:
        self.validate_node(node, code, parser, expected_position=position)

    @data_provider(
        (
            (
                lambda: cst.Integer("5", lpar=(cst.LeftParen(),)),
                "left paren without right paren",
            ),
            (
                lambda: cst.Integer("5", rpar=(cst.RightParen(),)),
                "right paren without left paren",
            ),
            (
                lambda: cst.Float("5.5", lpar=(cst.LeftParen(),)),
                "left paren without right paren",
            ),
            (
                lambda: cst.Float("5.5", rpar=(cst.RightParen(),)),
                "right paren without left paren",
            ),
            (
                lambda: cst.Imaginary("5i", lpar=(cst.LeftParen(),)),
                "left paren without right paren",
            ),
            (
                lambda: cst.Imaginary("5i", rpar=(cst.RightParen(),)),
                "right paren without left paren",
            ),
        )
    )
    def test_invalid(
        self, get_node: Callable[[], cst.CSTNode], expected_re: str
    ) -> None:
        self.assert_invalid(get_node, expected_re)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Callable

from libcst._parser.types.config import PartialParserConfig
from libcst.testing.utils import data_provider, UnitTest


class TestConfig(UnitTest):
    @data_provider(
        {
            "empty": (PartialParserConfig,),
            "python_version_a": (lambda: PartialParserConfig(python_version="3.7"),),
            "python_version_b": (lambda: PartialParserConfig(python_version="3.7.1"),),
            "encoding": (lambda: PartialParserConfig(encoding="latin-1"),),
            "default_indent": (lambda: PartialParserConfig(default_indent="\t    "),),
            "default_newline": (lambda: PartialParserConfig(default_newline="\r\n"),),
        }
    )
    def test_valid_partial_parser_config(
        self, factory: Callable[[], PartialParserConfig]
    ) -> None:
        self.assertIsInstance(factory(), PartialParserConfig)

    @data_provider(
        {
            "python_version": (
                lambda: PartialParserConfig(python_version="3.7.1.0"),
                "The given version is not in the right format",
            ),
            "python_version_unsupported": (
                lambda: PartialParserConfig(python_version="3.4"),
                "LibCST can only parse code using one of the following versions of Python's grammar",
            ),
            "encoding": (
                lambda: PartialParserConfig(encoding="utf-42"),
                "not a supported encoding",
            ),
            "default_indent": (
                lambda: PartialParserConfig(default_indent="badinput"),
                "invalid value for default_indent",
            ),
            "default_newline": (
                lambda: PartialParserConfig(default_newline="\n\r"),
                "invalid value for default_newline",
            ),
        }
    )
    def test_invalid_partial_parser_config(
        self, factory: Callable[[], PartialParserConfig], expected_re: str
    ) -> None:
        with self.assertRaisesRegex(ValueError, expected_re):
            factory()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any

import libcst as cst
from libcst import parse_statement
from libcst._nodes.tests.base import CSTNodeTest
from libcst.metadata import CodeRange
from libcst.testing.utils import data_provider


class AssignTest(CSTNodeTest):
    @data_provider(
        (
            # Simple assignment creation case.
            {
                "node": cst.Assign(
                    (cst.AssignTarget(cst.Name("foo")),), cst.Integer("5")
                ),
                "code": "foo = 5",
                "parser": None,
                "expected_position": CodeRange((1, 0), (1, 7)),
            },
            # Multiple targets creation
            {
                "node": cst.Assign(
                    (
                        cst.AssignTarget(cst.Name("foo")),
                        cst.AssignTarget(cst.Name("bar")),
                    ),
                    cst.Integer("5"),
                ),
                "code": "foo = bar = 5",
                "parser": None,
                "expected_position": CodeRange((1, 0), (1, 13)),
            },
            # Whitespace test for creating nodes
            {
                "node": cst.Assign(
                    (
                        cst.AssignTarget(
                            cst.Name("foo"),
                            whitespace_before_equal=cst.SimpleWhitespace(""),
                            whitespace_after_equal=cst.SimpleWhitespace(""),
                        ),
                    ),
                    cst.Integer("5"),
                ),
                "code": "foo=5",
                "parser": None,
                "expected_position": CodeRange((1, 0), (1, 5)),
            },
            # Simple assignment parser case.
            {
                "node": cst.SimpleStatementLine(
                    (
                        cst.Assign(
                            (cst.AssignTarget(cst.Name("foo")),), cst.Integer("5")
                        ),
                    )
                ),
                "code": "foo = 5\n",
                "parser": parse_statement,
                "expected_position": None,
            },
            # Multiple targets parser
            {
                "node": cst.SimpleStatementLine(
                    (
                        cst.Assign(
                            (
                                cst.AssignTarget(cst.Name("foo")),
                                cst.AssignTarget(cst.Name("bar")),
                            ),
                            cst.Integer("5"),
                        ),
                    )
                ),
                "code": "foo = bar = 5\n",
                "parser": parse_statement,
                "expected_position": None,
            },
            # Whitespace test parser
            {
                "node": cst.SimpleStatementLine(
                    (
                        cst.Assign(
                            (
                                cst.AssignTarget(
                                    cst.Name("foo"),
                                    whitespace_before_equal=cst.SimpleWhitespace(""),
                                    whitespace_after_equal=cst.SimpleWhitespace(""),
                                ),
                            ),
                            cst.Integer("5"),
                        ),
                    )
                ),
                "code": "foo=5\n",
                "parser": parse_statement,
                "expected_position": None,
            },
        )
    )
    def test_valid(self, **kwargs: Any) -> None:
        self.validate_node(**kwargs)

    @data_provider(
        (
            {
                "get_node": (lambda: cst.Assign(targets=(), value=cst.Integer("5"))),
                "expected_re": "at least one AssignTarget",
            },
        )
    )
    def test_invalid(self, **kwargs: Any) -> None:
        self.assert_invalid(**kwargs)

    @data_provider(
        (
            {
                "get_node": (
                    lambda: cst.Assign(
                        # pyre-ignore: Incompatible parameter type [6]
                        targets=[
                            cst.BinaryOperation(
                                left=cst.Name("x"),
                                operator=cst.Add(),
                                right=cst.Integer("1"),
                            ),
                        ],
                        value=cst.Name("y"),
                    )
                ),
                "expected_re": "Expected an instance of .*statement.AssignTarget.*",
            },
        )
    )
    def test_invalid_types(self, **kwargs: Any) -> None:
        self.assert_invalid_types(**kwargs)


class AnnAssignTest(CSTNodeTest):
    @data_provider(
        (
            # Simple assignment creation case.
            {
                "node": cst.AnnAssign(
                    cst.Name("foo"), cst.Annotation(cst.Name("str")), cst.Integer("5")
                ),
                "code": "foo: str = 5",
                "parser": None,
                "expected_position": CodeRange((1, 0), (1, 12)),
            },
            # Annotation creation without assignment
            {
                "node": cst.AnnAssign(cst.Name("foo"), cst.Annotation(cst.Name("str"))),
                "code": "foo: str",
                "parser": None,
                "expected_position": CodeRange((1, 0), (1, 8)),
            },
            # Complex annotation creation
            {
                "node": cst.AnnAssign(
                    cst.Name("foo"),
                    cst.Annotation(
                        cst.Subscript(
                            cst.Name("Optional"),
                            (cst.SubscriptElement(cst.Index(cst.Name("str"))),),
                        )
                    ),
                    cst.Integer("5"),
                ),
                "code": "foo: Optional[str] = 5",
                "parser": None,
                "expected_position": CodeRange((1, 0), (1, 22)),
            },
            # Simple assignment parser case.
            {
                "node": cst.SimpleStatementLine(
                    (
                        cst.AnnAssign(
                            target=cst.Name("foo"),
                            annotation=cst.Annotation(
                                annotation=cst.Name("str"),
                                whitespace_before_indicator=cst.SimpleWhitespace(""),
                            ),
                            equal=cst.AssignEqual(),
                            value=cst.Integer("5"),
                        ),
                    )
                ),
                "code": "foo: str = 5\n",
                "parser": parse_statement,
                "expected_position": None,
            },
            # Annotation without assignment
            {
                "node": cst.SimpleStatementLine(
                    (
                        cst.AnnAssign(
                            target=cst.Name("foo"),
                            annotation=cst.Annotation(
                                annotation=cst.Name("str"),
                                whitespace_before_indicator=cst.SimpleWhitespace(""),
                            ),
                            value=None,
                        ),
                    )
                ),
                "code": "foo: str\n",
                "parser": parse_statement,
                "expected_position": None,
            },
            # Complex annotation
            {
                "node": cst.SimpleStatementLine(
                    (
                        cst.AnnAssign(
                            target=cst.Name("foo"),
                            annotation=cst.Annotation(
                                annotation=cst.Subscript(
                                    cst.Name("Optional"),
                                    (cst.SubscriptElement(cst.Index(cst.Name("str"))),),
                                ),
                                whitespace_before_indicator=cst.SimpleWhitespace(""),
                            ),
                            equal=cst.AssignEqual(),
                            value=cst.Integer("5"),
                        ),
                    )
                ),
                "code": "foo: Optional[str] = 5\n",
                "parser": parse_statement,
                "expected_position": None,
            },
            # Whitespace test
            {
                "node": cst.AnnAssign(
                    target=cst.Name("foo"),
                    annotation=cst.Annotation(
                        annotation=cst.Subscript(
                            cst.Name("Optional"),
                            (cst.SubscriptElement(cst.Index(cst.Name("str"))),),
                        ),
                        whitespace_before_indicator=cst.SimpleWhitespace(" "),
                        whitespace_after_indicator=cst.SimpleWhitespace("  "),
                    ),
                    equal=cst.AssignEqual(
                        whitespace_before=cst.SimpleWhitespace("  "),
                        whitespace_after=cst.SimpleWhitespace("  "),
                    ),
                    value=cst.Integer("5"),
                ),
                "code": "foo :  Optional[str]  =  5",
                "parser": None,
                "expected_position": CodeRange((1, 0), (1, 26)),
            },
            {
                "node": cst.SimpleStatementLine(
                    (
                        cst.AnnAssign(
                            target=cst.Name("foo"),
                            annotation=cst.Annotation(
                                annotation=cst.Subscript(
                                    cst.Name("Optional"),
                                    (cst.SubscriptElement(cst.Index(cst.Name("str"))),),
                                ),
                                whitespace_before_indicator=cst.SimpleWhitespace(" "),
                                whitespace_after_indicator=cst.SimpleWhitespace("  "),
                            ),
                            equal=cst.AssignEqual(
                                whitespace_before=cst.SimpleWhitespace("  "),
                                whitespace_after=cst.SimpleWhitespace("  "),
                            ),
                            value=cst.Integer("5"),
                        ),
                    )
                ),
                "code": "foo :  Optional[str]  =  5\n",
                "parser": parse_statement,
                "expected_position": None,
            },
        )
    )
    def test_valid(self, **kwargs: Any) -> None:
        self.validate_node(**kwargs)

    @data_provider(
        (
            {
                "get_node": (
                    lambda: cst.AnnAssign(
                        target=cst.Name("foo"),
                        annotation=cst.Annotation(cst.Name("str")),
                        equal=cst.AssignEqual(),
                        value=None,
                    )
                ),
                "expected_re": "Must have a value when specifying an AssignEqual.",
            },
        )
    )
    def test_invalid(self, **kwargs: Any) -> None:
        self.assert_invalid(**kwargs)

    @data_provider(
        (
            {
                "get_node": (
                    lambda: cst.AnnAssign(
                        # pyre-ignore: Incompatible parameter type [6]
                        target=cst.BinaryOperation(
                            left=cst.Name("x"),
                            operator=cst.Add(),
                            right=cst.Integer("1"),
                        ),
                        annotation=cst.Annotation(cst.Name("int")),
                        equal=cst.AssignEqual(),
                        value=cst.Name("y"),
                    )
                ),
                "expected_re": (
                    "Expected an instance of .*BaseAssignTargetExpression.*"
                ),
            },
        )
    )
    def test_invalid_types(self, **kwargs: Any) -> None:
        self.assert_invalid_types(**kwargs)


class AugAssignTest(CSTNodeTest):
    @data_provider(
        (
            # Simple assignment constructor case.
            {
                "node": cst.AugAssign(
                    cst.Name("foo"), cst.AddAssign(), cst.Integer("5")
                ),
                "code": "foo += 5",
                "parser": None,
                "expected_position": CodeRange((1, 0), (1, 8)),
            },
            {
                "node": cst.AugAssign(
                    cst.Name("bar"), cst.MultiplyAssign(), cst.Name("foo")
                ),
                "code": "bar *= foo",
                "parser": None,
                "expected_position": None,
            },
            # Whitespace constructor test
            {
                "node": cst.AugAssign(
                    target=cst.Name("foo"),
                    operator=cst.LeftShiftAssign(
                        whitespace_before=cst.SimpleWhitespace("  "),
                        whitespace_after=cst.SimpleWhitespace("  "),
                    ),
                    value=cst.Integer("5"),
                ),
                "code": "foo  <<=  5",
                "parser": None,
                "expected_position": CodeRange((1, 0), (1, 11)),
            },
            # Simple assignment parser case.
            {
                "node": cst.SimpleStatementLine(
                    (cst.AugAssign(cst.Name("foo"), cst.AddAssign(), cst.Integer("5")),)
                ),
                "code": "foo += 5\n",
                "parser": parse_statement,
                "expected_position": None,
            },
            {
                "node": cst.SimpleStatementLine(
                    (
                        cst.AugAssign(
                            cst.Name("bar"), cst.MultiplyAssign(), cst.Name("foo")
                        ),
                    )
                ),
                "code": "bar *= foo\n",
                "parser": parse_statement,
                "expected_position": None,
            },
            # Whitespace parser test
            {
                "node": cst.SimpleStatementLine(
                    (
                        cst.AugAssign(
                            target=cst.Name("foo"),
                            operator=cst.LeftShiftAssign(
                                whitespace_before=cst.SimpleWhitespace("  "),
                                whitespace_after=cst.SimpleWhitespace("  "),
                            ),
                            value=cst.Integer("5"),
                        ),
                    )
                ),
                "code": "foo  <<=  5\n",
                "parser": parse_statement,
                "expected_position": None,
            },
        )
    )
    def test_valid(self, **kwargs: Any) -> None:
        self.validate_node(**kwargs)

    @data_provider(
        (
            {
                "get_node": (
                    lambda: cst.AugAssign(
                        # pyre-ignore: Incompatible parameter type [6]
                        target=cst.BinaryOperation(
                            left=cst.Name("x"),
                            operator=cst.Add(),
                            right=cst.Integer("1"),
                        ),
                        operator=cst.AddAssign(),
                        value=cst.Name("y"),
                    )
                ),
                "expected_re": (
                    "Expected an instance of .*BaseAssignTargetExpression.*"
                ),
            },
        )
    )
    def test_invalid_types(self, **kwargs: Any) -> None:
        self.assert_invalid_types(**kwargs)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


from typing import Optional

import libcst as cst
from libcst.metadata.base_provider import BatchableMetadataProvider


class ParentNodeVisitor(cst.CSTVisitor):
    def __init__(self, provider: "ParentNodeProvider") -> None:
        self.provider: ParentNodeProvider = provider
        super().__init__()

    def on_leave(self, original_node: cst.CSTNode) -> None:
        for child in original_node.children:
            self.provider.set_metadata(child, original_node)
        super().on_leave(original_node)


class ParentNodeProvider(BatchableMetadataProvider[cst.CSTNode]):
    def visit_Module(self, node: cst.Module) -> Optional[bool]:
        node.visit(ParentNodeVisitor(self))



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Callable, Optional

import libcst as cst
from libcst import parse_statement
from libcst._nodes.tests.base import CSTNodeTest
from libcst._parser.entrypoints import is_native
from libcst.testing.utils import data_provider

parser: Optional[Callable[[str], cst.CSTNode]] = (
    parse_statement if is_native() else None
)


class MatchTest(CSTNodeTest):
    @data_provider(
        (
            # Values and singletons
            {
                "node": cst.Match(
                    subject=cst.Name("x"),
                    cases=[
                        cst.MatchCase(
                            pattern=cst.MatchSingleton(cst.Name("None")),
                            body=cst.SimpleStatementSuite((cst.Pass(),)),
                        ),
                        cst.MatchCase(
                            pattern=cst.MatchValue(cst.SimpleString('"foo"')),
                            body=cst.SimpleStatementSuite((cst.Pass(),)),
                        ),
                    ],
                ),
                "code": "match x:\n"
                + "    case None: pass\n"
                + '    case "foo": pass\n',
                "parser": parser,
            },
            # Parenthesized value
            {
                "node": cst.Match(
                    subject=cst.Name(
                        value="x",
                    ),
                    cases=[
                        cst.MatchCase(
                            pattern=cst.MatchAs(
                                pattern=cst.MatchValue(
                                    value=cst.Integer(
                                        value="1",
                                        lpar=[
                                            cst.LeftParen(),
                                        ],
                                        rpar=[
                                            cst.RightParen(),
                                        ],
                                    ),
                                ),
                                name=cst.Name(
                                    value="z",
                                ),
                                whitespace_before_as=cst.SimpleWhitespace(" "),
                                whitespace_after_as=cst.SimpleWhitespace(" "),
                            ),
                            body=cst.SimpleStatementSuite([cst.Pass()]),
                        ),
                    ],
                ),
                "code": "match x:\n    case (1) as z: pass\n",
                "parser": parser,
            },
            # List patterns
            {
                "node": cst.Match(
                    subject=cst.Name("x"),
                    cases=[
                        cst.MatchCase(  # empty list
                            pattern=cst.MatchList(
                                [],
                                lbracket=cst.LeftSquareBracket(),
                                rbracket=cst.RightSquareBracket(),
                            ),
                            body=cst.SimpleStatementSuite((cst.Pass(),)),
                        ),
                        cst.MatchCase(  # single element list
                            pattern=cst.MatchList(
                                [
                                    cst.MatchSequenceElement(
                                        cst.MatchSingleton(cst.Name("None"))
                                    )
                                ],
                                lbracket=cst.LeftSquareBracket(),
                                rbracket=cst.RightSquareBracket(),
                            ),
                            body=cst.SimpleStatementSuite((cst.Pass(),)),
                        ),
                        cst.MatchCase(  # single element list with trailing comma
                            pattern=cst.MatchList(
                                [
                                    cst.MatchSequenceElement(
                                        cst.MatchSingleton(cst.Name("None")),
                                        cst.Comma(),
                                    )
                                ],
                                lbracket=cst.LeftSquareBracket(),
                                rbracket=cst.RightSquareBracket(),
                            ),
                            body=cst.SimpleStatementSuite((cst.Pass(),)),
                        ),
                    ],
                ),
                "code": (
                    "match x:\n"
                    + "    case []: pass\n"
                    + "    case [None]: pass\n"
                    + "    case [None,]: pass\n"
                ),
                "parser": parser,
            },
            # Tuple patterns
            {
                "node": cst.Match(
                    subject=cst.Name("x"),
                    cases=[
                        cst.MatchCase(  # empty tuple
                            pattern=cst.MatchTuple(
                                [],
                            ),
                            body=cst.SimpleStatementSuite((cst.Pass(),)),
                        ),
                        cst.MatchCase(  # two element tuple
                            pattern=cst.MatchTuple(
                                [
                                    cst.MatchSequenceElement(
                                        cst.MatchSingleton(cst.Name("None")),
                                        cst.Comma(),
                                    ),
                                    cst.MatchSequenceElement(
                                        cst.MatchSingleton(cst.Name("None")),
                                    ),
                                ],
                            ),
                            body=cst.SimpleStatementSuite((cst.Pass(),)),
                        ),
                        cst.MatchCase(  # single element tuple with trailing comma
                            pattern=cst.MatchTuple(
                                [
                                    cst.MatchSequenceElement(
                                        cst.MatchSingleton(cst.Name("None")),
                                        cst.Comma(),
                                    )
                                ],
                            ),
                            body=cst.SimpleStatementSuite((cst.Pass(),)),
                        ),
                        cst.MatchCase(  # two element tuple
                            pattern=cst.MatchTuple(
                                [
                                    cst.MatchSequenceElement(
                                        cst.MatchSingleton(cst.Name("None")),
                                        cst.Comma(),
                                    ),
                                    cst.MatchStar(
                                        comma=cst.Comma(),
                                    ),
                                    cst.MatchSequenceElement(
                                        cst.MatchSingleton(cst.Name("None")),
                                    ),
                                ],
                            ),
                            body=cst.SimpleStatementSuite((cst.Pass(),)),
                        ),
                    ],
                ),
                "code": (
                    "match x:\n"
                    + "    case (): pass\n"
                    + "    case (None,None): pass\n"
                    + "    case (None,): pass\n"
                    + "    case (None,*_,None): pass\n"
                ),
                "parser": parser,
            },
            # Mapping patterns
            {
                "node": cst.Match(
                    subject=cst.Name("x"),
                    cases=[
                        cst.MatchCase(  # empty mapping
                            pattern=cst.MatchMapping(
                                [],
                            ),
                            body=cst.SimpleStatementSuite((cst.Pass(),)),
                        ),
                        cst.MatchCase(  # two element mapping
                            pattern=cst.MatchMapping(
                                [
                                    cst.MatchMappingElement(
                                        key=cst.SimpleString('"a"'),
                                        pattern=cst.MatchSingleton(cst.Name("None")),
                                        comma=cst.Comma(),
                                    ),
                                    cst.MatchMappingElement(
                                        key=cst.SimpleString('"b"'),
                                        pattern=cst.MatchSingleton(cst.Name("None")),
                                    ),
                                ],
                            ),
                            body=cst.SimpleStatementSuite((cst.Pass(),)),
                        ),
                        cst.MatchCase(  # single element mapping with trailing comma
                            pattern=cst.MatchMapping(
                                [
                                    cst.MatchMappingElement(
                                        key=cst.SimpleString('"a"'),
                                        pattern=cst.MatchSingleton(cst.Name("None")),
                                        comma=cst.Comma(),
                                    )
                                ],
                            ),
                            body=cst.SimpleStatementSuite((cst.Pass(),)),
                        ),
                        cst.MatchCase(  # rest
                            pattern=cst.MatchMapping(
                                rest=cst.Name("rest"),
                            ),
                            body=cst.SimpleStatementSuite((cst.Pass(),)),
                        ),
                    ],
                ),
                "code": (
                    "match x:\n"
                    + "    case {}: pass\n"
                    + '    case {"a": None,"b": None}: pass\n'
                    + '    case {"a": None,}: pass\n'
                    + "    case {**rest}: pass\n"
                ),
                "parser": parser,
            },
            # Class patterns
            {
                "node": cst.Match(
                    subject=cst.Name("x"),
                    cases=[
                        cst.MatchCase(  # empty class
                            pattern=cst.MatchClass(
                                cls=cst.Attribute(cst.Name("a"), cst.Name("b")),
                            ),
                            body=cst.SimpleStatementSuite((cst.Pass(),)),
                        ),
                        cst.MatchCase(  # single pattern class
                            pattern=cst.MatchClass(
                                cls=cst.Attribute(cst.Name("a"), cst.Name("b")),
                                patterns=[
                                    cst.MatchSequenceElement(
                                        cst.MatchSingleton(cst.Name("None"))
                                    )
                                ],
                            ),
                            body=cst.SimpleStatementSuite((cst.Pass(),)),
                        ),
                        cst.MatchCase(  # single pattern class with trailing comma
                            pattern=cst.MatchClass(
                                cls=cst.Attribute(cst.Name("a"), cst.Name("b")),
                                patterns=[
                                    cst.MatchSequenceElement(
                                        cst.MatchSingleton(cst.Name("None")),
                                        comma=cst.Comma(),
                                    )
                                ],
                            ),
                            body=cst.SimpleStatementSuite((cst.Pass(),)),
                        ),
                        cst.MatchCase(  # single keyword pattern class
                            pattern=cst.MatchClass(
                                cls=cst.Attribute(cst.Name("a"), cst.Name("b")),
                                kwds=[
                                    cst.MatchKeywordElement(
                                        key=cst.Name("foo"),
                                        pattern=cst.MatchSingleton(cst.Name("None")),
                                    )
                                ],
                            ),
                            body=cst.SimpleStatementSuite((cst.Pass(),)),
                        ),
                        cst.MatchCase(  # single keyword pattern class with trailing comma
                            pattern=cst.MatchClass(
                                cls=cst.Attribute(cst.Name("a"), cst.Name("b")),
                                kwds=[
                                    cst.MatchKeywordElement(
                                        key=cst.Name("foo"),
                                        pattern=cst.MatchSingleton(cst.Name("None")),
                                        comma=cst.Comma(),
                                    )
                                ],
                            ),
                            body=cst.SimpleStatementSuite((cst.Pass(),)),
                        ),
                        cst.MatchCase(  # now all at once
                            pattern=cst.MatchClass(
                                cls=cst.Attribute(cst.Name("a"), cst.Name("b")),
                                patterns=[
                                    cst.MatchSequenceElement(
                                        cst.MatchSingleton(cst.Name("None")),
                                        cst.Comma(),
                                    ),
                                    cst.MatchSequenceElement(
                                        cst.MatchSingleton(cst.Name("None")),
                                        cst.Comma(),
                                    ),
                                ],
                                kwds=[
                                    cst.MatchKeywordElement(
                                        key=cst.Name("foo"),
                                        pattern=cst.MatchSingleton(cst.Name("None")),
                                        comma=cst.Comma(),
                                    ),
                                    cst.MatchKeywordElement(
                                        key=cst.Name("bar"),
                                        pattern=cst.MatchSingleton(cst.Name("None")),
                                        comma=cst.Comma(),
                                    ),
                                ],
                            ),
                            body=cst.SimpleStatementSuite((cst.Pass(),)),
                        ),
                    ],
                ),
                "code": (
                    "match x:\n"
                    + "    case a.b(): pass\n"
                    + "    case a.b(None): pass\n"
                    + "    case a.b(None,): pass\n"
                    + "    case a.b(foo=None): pass\n"
                    + "    case a.b(foo=None,): pass\n"
                    + "    case a.b(None,None,foo=None,bar=None,): pass\n"
                ),
                "parser": parser,
            },
            # as pattern
            {
                "node": cst.Match(
                    subject=cst.Name("x"),
                    cases=[
                        cst.MatchCase(
                            pattern=cst.MatchAs(),
                            body=cst.SimpleStatementSuite((cst.Pass(),)),
                        ),
                        cst.MatchCase(
                            pattern=cst.MatchAs(name=cst.Name("foo")),
                            body=cst.SimpleStatementSuite((cst.Pass(),)),
                        ),
                        cst.MatchCase(
                            pattern=cst.MatchAs(
                                pattern=cst.MatchSingleton(cst.Name("None")),
                                name=cst.Name("bar"),
                                whitespace_before_as=cst.SimpleWhitespace(" "),
                                whitespace_after_as=cst.SimpleWhitespace(" "),
                            ),
                            body=cst.SimpleStatementSuite((cst.Pass(),)),
                        ),
                    ],
                ),
                "code": "match x:\n"
                + "    case _: pass\n"
                + "    case foo: pass\n"
                + "    case None as bar: pass\n",
                "parser": parser,
            },
            # or pattern
            {
                "node": cst.Match(
                    subject=cst.Name("x"),
                    cases=[
                        cst.MatchCase(
                            pattern=cst.MatchOr(
                                [
                                    cst.MatchOrElement(
                                        cst.MatchSingleton(cst.Name("None")),
                                        cst.BitOr(),
                                    ),
                                    cst.MatchOrElement(
                                        cst.MatchSingleton(cst.Name("False")),
                                        cst.BitOr(),
                                    ),
                                    cst.MatchOrElement(
                                        cst.MatchSingleton(cst.Name("True"))
                                    ),
                                ]
                            ),
                            body=cst.SimpleStatementSuite((cst.Pass(),)),
                        )
                    ],
                ),
                "code": "match x:\n    case None | False | True: pass\n",
                "parser": parser,
            },
            {  # exercise sentinels
                "node": cst.Match(
                    subject=cst.Name("x"),
                    cases=[
                        cst.MatchCase(
                            pattern=cst.MatchList(
                                [cst.MatchStar(), cst.MatchStar()],
                                lbracket=None,
                                rbracket=None,
                            ),
                            body=cst.SimpleStatementSuite((cst.Pass(),)),
                        ),
                        cst.MatchCase(
                            pattern=cst.MatchTuple(
                                [
                                    cst.MatchSequenceElement(
                                        cst.MatchSingleton(cst.Name("None"))
                                    )
                                ]
                            ),
                            body=cst.SimpleStatementSuite((cst.Pass(),)),
                        ),
                        cst.MatchCase(
                            pattern=cst.MatchAs(
                                pattern=cst.MatchTuple(
                                    [
                                        cst.MatchSequenceElement(
                                            cst.MatchSingleton(cst.Name("None"))
                                        )
                                    ]
                                ),
                                name=cst.Name("bar"),
                            ),
                            body=cst.SimpleStatementSuite((cst.Pass(),)),
                        ),
                        cst.MatchCase(
                            pattern=cst.MatchOr(
                                [
                                    cst.MatchOrElement(
                                        cst.MatchSingleton(cst.Name("None")),
                                    ),
                                    cst.MatchOrElement(
                                        cst.MatchSingleton(cst.Name("False")),
                                    ),
                                    cst.MatchOrElement(
                                        cst.MatchSingleton(cst.Name("True"))
                                    ),
                                ]
                            ),
                            body=cst.SimpleStatementSuite((cst.Pass(),)),
                        ),
                    ],
                ),
                "code": "match x:\n"
                + "    case *_, *_: pass\n"
                + "    case (None,): pass\n"
                + "    case (None,) as bar: pass\n"
                + "    case None | False | True: pass\n",
                "parser": None,
            },
            # Match without whitespace between keyword and the expr
            {
                "node": cst.Match(
                    subject=cst.Name(
                        "x", lpar=[cst.LeftParen()], rpar=[cst.RightParen()]
                    ),
                    cases=[
                        cst.MatchCase(
                            pattern=cst.MatchSingleton(
                                cst.Name(
                                    "None",
                                    lpar=[cst.LeftParen()],
                                    rpar=[cst.RightParen()],
                                )
                            ),
                            body=cst.SimpleStatementSuite((cst.Pass(),)),
                            whitespace_after_case=cst.SimpleWhitespace(
                                value="",
                            ),
                        ),
                    ],
                    whitespace_after_match=cst.SimpleWhitespace(
                        value="",
                    ),
                ),
                "code": "match(x):\n    case(None): pass\n",
                "parser": parser,
            },
        )
    )
    def test_valid(self, **kwargs: Any) -> None:
        self.validate_node(**kwargs)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
from typing import Mapping

import libcst as cst
from libcst import MetadataDependent
from libcst.codemod._codemod import Codemod
from libcst.codemod._context import CodemodContext
from libcst.matchers import MatcherDecoratableTransformer, MatcherDecoratableVisitor
from libcst.metadata import ProviderT


class ContextAwareTransformer(Codemod, MatcherDecoratableTransformer):
    """
    A transformer which visits using LibCST. Allows visitor-based mutation of a tree.
    Classes wishing to do arbitrary non-visitor-based mutation on a tree should
    instead subclass from :class:`Codemod` and implement
    :meth:`~Codemod.transform_module_impl`. This is a subclass of
    :class:`~libcst.matchers.MatcherDecoratableTransformer` so all features of matchers
    as well as :class:`~libcst.CSTTransformer` are available to subclasses of this
    class.
    """

    def __init__(self, context: CodemodContext) -> None:
        Codemod.__init__(self, context)
        MatcherDecoratableTransformer.__init__(self)

    def transform_module_impl(self, tree: cst.Module) -> cst.Module:
        return tree.visit(self)


class ContextAwareVisitor(MatcherDecoratableVisitor, MetadataDependent):
    """
    A visitor which visits using LibCST. Allows visitor-based collecting of info
    on a tree. All codemods which wish to implement an information collector should
    subclass from this instead of directly from
    :class:`~libcst.matchers.MatcherDecoratableVisitor` or :class:`~libcst.CSTVisitor`
    since this provides access to the current codemod context. As a result, this
    class allows access to metadata which was calculated in a parent
    :class:`~libcst.codemod.Codemod` through the
    :meth:`~libcst.MetadataDependent.get_metadata` method.

    Note that you cannot directly run a :class:`~libcst.codemod.ContextAwareVisitor`
    using :func:`~libcst.codemod.transform_module` because visitors by definition
    do not transform trees. However, you can instantiate a
    :class:`~libcst.codemod.ContextAwareVisitor` inside a codemod and pass it to the
    :class:`~libcst.CSTNode.visit` method on any node in order to run information
    gathering with metadata and context support.

    Remember that a :class:`~libcst.codemod.ContextAwareVisitor` is a subclass of
    :class:`~libcst.MetadataDependent`, meaning that you still need to declare
    your metadata dependencies with
    :attr:`~libcst.MetadataDependent.METADATA_DEPENDENCIES` before you can retrieve
    metadata using :meth:`~libcst.MetadataDependent.get_metadata`, even if the parent
    codemod has listed its own metadata dependencies. Note also that the dependencies
    listed on this class must be a strict subset of the dependencies listed in the
    parent codemod.
    """

    def __init__(self, context: CodemodContext) -> None:
        MetadataDependent.__init__(self)
        MatcherDecoratableVisitor.__init__(self)
        self.context = context

        dependencies = self.get_inherited_dependencies()
        if dependencies:
            wrapper = self.context.wrapper
            if wrapper is None:
                raise Exception(
                    f"Attempting to instantiate {self.__class__.__name__} outside of "
                    + "an active transform. This means that metadata hasn't been "
                    + "calculated and we cannot successfully create this visitor."
                )
            for dep in dependencies:
                if dep not in wrapper._metadata:
                    raise Exception(
                        f"Attempting to access metadata {dep.__name__} that was not a "
                        + "declared dependency of parent transform! This means it is "
                        + "not possible to compute this value. Please ensure that all "
                        + f"parent transforms of {self.__class__.__name__} declare "
                        + f"{dep.__name__} as a metadata dependency."
                    )
            self.metadata: Mapping[ProviderT, Mapping[cst.CSTNode, object]] = {
                dep: wrapper._metadata[dep] for dep in dependencies
            }

    def warn(self, warning: str) -> None:
        """
        Emit a warning that is displayed to the user who has invoked this codemod.
        """
        self.context.warnings.append(warning)

    @property
    def module(self) -> cst.Module:
        """
        Reference to the currently-traversed module. Note that this is only available
        during a transform itself.
        """
        module = self.context.module
        if module is None:
            raise Exception(
                f"Attempted access of {self.__class__.__name__}.module outside of "
                + "transform_module()."
            )
        return module



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

class TokenType:
    name: str
    contains_syntax: bool

STRING: TokenType = ...
NAME: TokenType = ...
NUMBER: TokenType = ...
OP: TokenType = ...
NEWLINE: TokenType = ...
INDENT: TokenType = ...
DEDENT: TokenType = ...
ASYNC: TokenType = ...
AWAIT: TokenType = ...
FSTRING_START: TokenType = ...
FSTRING_STRING: TokenType = ...
FSTRING_END: TokenType = ...
ENDMARKER: TokenType = ...
# unused dummy tokens for backwards compat with the parso tokenizer
ERRORTOKEN: TokenType = ...
ERROR_DEDENT: TokenType = ...



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Type, Union

import libcst as cst
from libcst import parse_module, RemovalSentinel
from libcst._nodes.tests.base import CSTNodeTest
from libcst._types import CSTNodeT
from libcst._visitors import CSTTransformer
from libcst.testing.utils import data_provider


class IfStatementRemovalVisitor(CSTTransformer):
    def on_leave(
        self, original_node: CSTNodeT, updated_node: CSTNodeT
    ) -> Union[CSTNodeT, RemovalSentinel]:
        if isinstance(updated_node, cst.If):
            return cst.RemoveFromParent()
        else:
            return updated_node


class ContinueStatementRemovalVisitor(CSTTransformer):
    def on_leave(
        self, original_node: CSTNodeT, updated_node: CSTNodeT
    ) -> Union[CSTNodeT, RemovalSentinel]:
        if isinstance(updated_node, cst.Continue):
            return cst.RemoveFromParent()
        else:
            return updated_node


class SpecificImportRemovalVisitor(CSTTransformer):
    def on_leave(
        self, original_node: CSTNodeT, updated_node: CSTNodeT
    ) -> Union[cst.Import, cst.ImportFrom, CSTNodeT, RemovalSentinel]:
        if isinstance(updated_node, cst.Import):
            for alias in updated_node.names:
                name = alias.name
                if isinstance(name, cst.Name) and name.value == "b":
                    return cst.RemoveFromParent()
        elif isinstance(updated_node, cst.ImportFrom):
            module = updated_node.module
            if isinstance(module, cst.Name) and module.value == "e":
                return cst.RemoveFromParent()
        return updated_node


class RemovalBehavior(CSTNodeTest):
    @data_provider(
        (
            # Top of module doesn't require a pass, empty code is valid.
            ("continue", "", ContinueStatementRemovalVisitor),
            ("if condition: print('hello world')", "", IfStatementRemovalVisitor),
            # Verify behavior within an indented block.
            (
                "while True:\n    continue",
                "while True:\n    pass",
                ContinueStatementRemovalVisitor,
            ),
            (
                "while True:\n    if condition: print('hello world')",
                "while True:\n    pass",
                IfStatementRemovalVisitor,
            ),
            # Verify behavior within a simple statement suite.
            (
                "while True: continue",
                "while True: pass",
                ContinueStatementRemovalVisitor,
            ),
            # Verify with some imports
            (
                "import a\nimport b\n\nfrom c import d\nfrom e import f",
                "import a\n\nfrom c import d",
                SpecificImportRemovalVisitor,
            ),
            # Verify only one pass is generated even if we remove multiple statements
            (
                "while True:\n    continue\ncontinue",
                "while True:\n    pass",
                ContinueStatementRemovalVisitor,
            ),
            (
                "while True: continue ; continue",
                "while True: pass",
                ContinueStatementRemovalVisitor,
            ),
        )
    )
    def test_removal_pass_behavior(
        self, before: str, after: str, visitor: Type[CSTTransformer]
    ) -> None:
        if before.endswith("\n") or after.endswith("\n"):
            raise Exception("Test cases should not be newline-terminated!")

        # Test doesn't have newline termination case
        before_module = parse_module(before)
        after_module = before_module.visit(visitor())
        self.assertEqual(after, after_module.code)

        # Test does have newline termination case
        before_module = parse_module(before + "\n")
        after_module = before_module.visit(visitor())
        self.assertEqual(after + "\n", after_module.code)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Parso doesn't attempt to parse (or even emit tokens for) whitespace or comments that
aren't syntatically important. Instead, we're just given the whitespace as a "prefix" of
the token.

However, in our CST, whitespace is gathered into far more detailed objects than a simple
str.

Fortunately this isn't hard for us to parse ourselves, so we just use our own
hand-rolled recursive descent parser.
"""

try:
    # It'd be better to do `from libcst_native.whitespace_parser import *`, but we're
    # blocked on https://github.com/PyO3/pyo3/issues/759
    # (which ultimately seems to be a limitation of how importlib works)
    from libcst_native import whitespace_parser as mod
except ImportError:
    from libcst._parser import py_whitespace_parser as mod

parse_simple_whitespace = mod.parse_simple_whitespace
parse_empty_lines = mod.parse_empty_lines
parse_trailing_whitespace = mod.parse_trailing_whitespace
parse_parenthesizable_whitespace = mod.parse_parenthesizable_whitespace



# see https://github.com/python/cpython/pull/31018/files#diff-3f516b60719dd445d33225e4f316b36e85c9c51a843a0147349d11a005c55937

A[*b]
A[  *  b  ]
A[ *  b ,  ]
A[*b] = 1
del A[*b]

A[* b ,  * b]
A[ b, *b]
A[* b, b]
A[ *  b,b, b]
A[b, *b, b]

A[*A[b, *b, b], b]
A[b, ...]
A[*A[b, ...]]

A[ * ( 1,2,3)]
A[ * [ 1,2,3]]

A[1:2, *t]
A[1:, *t, 1:2]
A[:, *t, :]
A[*t, :, *t]

A[* returns_list()]
A[*returns_list(), * returns_list(), b]

def f1(*args: *b): pass
def f2(*args: *b, arg1): pass
def f3(*args: *b, arg1: int): pass
def f4(*args: *b, arg1: int = 1): pass

def f(*args: *tuple[int, ...]): pass
def f(*args: *tuple[int, *Ts]): pass
def f() -> tuple[int, *tuple[int, ...]]: pass


#foo.

try   :
    pass

    # foo

except *  lol  as  LOL  :

    pass

except  * f:

    # foo

    pass

else  :

    pass

finally  :

    foo

try:
    pass
except*f:
    pass
finally:
    pass


try:

    # 1

    try:

        #  2

        pass

        # 3

    # 4

    finally:

        # 5

        pass
    
        # 6
    
    # 7

except *foo:

    #8

    pass

    #9



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import re
from typing import Dict, Pattern, Union

import libcst as cst
from libcst.codemod._context import CodemodContext
from libcst.codemod._visitor import ContextAwareVisitor
from libcst.metadata import PositionProvider


class GatherCommentsVisitor(ContextAwareVisitor):
    """
    Collects all comments matching a certain regex and their line numbers.
    This visitor is useful for capturing special-purpose comments, for example
    ``noqa`` style lint suppression annotations.

    Standalone comments are assumed to affect the line following them, and
    inline ones are recorded with the line they are on.

    After visiting a CST, matching comments are collected in the ``comments``
    attribute.
    """

    METADATA_DEPENDENCIES = (PositionProvider,)

    def __init__(self, context: CodemodContext, comment_regex: str) -> None:
        super().__init__(context)

        #: Dictionary of comments found in the CST. Keys are line numbers,
        #: values are comment nodes.
        self.comments: Dict[int, cst.Comment] = {}

        self._comment_matcher: Pattern[str] = re.compile(comment_regex)

    def visit_EmptyLine(self, node: cst.EmptyLine) -> bool:
        if node.comment is not None:
            self.handle_comment(node)
        return False

    def visit_TrailingWhitespace(self, node: cst.TrailingWhitespace) -> bool:
        if node.comment is not None:
            self.handle_comment(node)
        return False

    def handle_comment(
        self, node: Union[cst.EmptyLine, cst.TrailingWhitespace]
    ) -> None:
        comment = node.comment
        assert comment is not None  # ensured by callsites above
        if not self._comment_matcher.match(comment.value):
            return
        line = self.get_metadata(PositionProvider, comment).start.line
        if isinstance(node, cst.EmptyLine):
            # Standalone comments refer to the next line
            line += 1
        self.comments[line] = comment



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
from typing import Type, TypeVar

T = TypeVar("T")


def ensure_type(node: object, nodetype: Type[T]) -> T:
    """
    Takes any python object, and a LibCST :class:`~libcst.CSTNode` subclass and
    refines the type of the python object. This is most useful when you already
    know that a particular object is a certain type but your type checker is not
    convinced. Note that this does an instance check for you and raises an
    exception if it is not the right type, so this should be used in situations
    where you are sure of the type given previous checks.
    """

    if not isinstance(node, nodetype):
        raise Exception(
            f"Expected a {nodetype.__name__} but got a {node.__class__.__name__}!"
        )
    return node
